import pathlib

import numpy
import pytest

from tidyms2.core import exceptions
from tidyms2.core.matrix import DataMatrix, FeatureVector, SampleVector, validate_data_matrix
from tidyms2.core.models import FeatureGroup, Sample

from ..helpers import create_feature_group, create_sample


def ft_transformer(data: FeatureVector, **kwargs):
    """Feature transformer function for tests."""
    data.data[:] = 1.0
    return data


def sample_transformer(data: SampleVector, **kwargs):
    """Feature transformer function for tests."""
    data.data[:] = 1.0
    return data


class TestValidateDataMatrix:
    def test_create_matrix_using_samples_with_repeated_ids_raise_error(self, tmp_path: pathlib.Path):
        samples = [create_sample(tmp_path, 1), create_sample(tmp_path, 1)]
        features = [create_feature_group(x) for x in range(5)]
        data = numpy.random.normal(loc=100.0, size=(2, 5))
        with pytest.raises(exceptions.RepeatedIdError):
            validate_data_matrix(samples, features, data)

    def test_create_matrix_using_samples_with_repeated_order_raise_error(self, tmp_path: pathlib.Path):
        s1 = create_sample(tmp_path, 1)
        s1.meta.order = 1
        s2 = create_sample(tmp_path, 2)
        s2.meta.order = 1
        samples = [s1, s2]
        features = [create_feature_group(x) for x in range(5)]
        data = numpy.random.normal(loc=100.0, size=(2, 5))
        with pytest.raises(exceptions.RepeatedSampleOrder):
            validate_data_matrix(samples, features, data)

    def test_create_matrix_using_features_with_repeated_group_raise_error(self, tmp_path: pathlib.Path):
        samples = [create_sample(tmp_path, 1), create_sample(tmp_path, 1)]
        features = [create_feature_group(x) for x in range(5)]
        features[2].group = 3
        features[3].group = 3
        data = numpy.random.normal(loc=100.0, size=(2, 5))
        with pytest.raises(exceptions.RepeatedIdError):
            validate_data_matrix(samples, features, data)

    def test_create_matrix_with_non_matching_sizes_in_sample_data_raises_error(self, tmp_path: pathlib.Path):
        samples = [create_sample(tmp_path, 1), create_sample(tmp_path, 1)]
        features = [create_feature_group(x) for x in range(5)]
        data = numpy.random.normal(loc=100.0, size=(3, 5))
        with pytest.raises(ValueError):
            validate_data_matrix(samples, features, data)

    def test_create_matrix_with_non_matching_sizes_in_feature_data_raises_error(self, tmp_path: pathlib.Path):
        samples = [create_sample(tmp_path, 1), create_sample(tmp_path, 1)]
        features = [create_feature_group(x) for x in range(5)]
        data = numpy.random.normal(loc=100.0, size=(2, 6))
        with pytest.raises(ValueError):
            validate_data_matrix(samples, features, data)

    def test_create_matrix_without_samples_raises_error(self):
        samples = list()
        features = [create_feature_group(x) for x in range(5)]
        data = numpy.random.normal(loc=100.0, size=(2, 5))
        with pytest.raises(exceptions.EmptyDataMatrix):
            validate_data_matrix(samples, features, data)

    def test_create_matrix_without_features_raises_error(self, tmp_path: pathlib.Path):
        samples = [create_sample(tmp_path, 1), create_sample(tmp_path, 1)]
        features = []
        data = numpy.random.normal(loc=100.0, size=(2, 5))
        with pytest.raises(exceptions.EmptyDataMatrix):
            validate_data_matrix(samples, features, data)

    def test_create_matrix_using_non_2d_array_shape_raises_error(self, tmp_path: pathlib.Path):
        samples = [create_sample(tmp_path, x) for x in range(3)]
        features = [create_feature_group(x) for x in range(5)]
        data = numpy.random.normal(loc=100.0, size=(3))
        with pytest.raises(ValueError):
            validate_data_matrix(samples, features, data)

    def test_create_matrix_using_non_float_array_shape_raises_error(self, tmp_path: pathlib.Path):
        samples = [create_sample(tmp_path, x) for x in range(3)]
        features = [create_feature_group(x) for x in range(5)]
        data = numpy.ones(shape=(3, 5), dtype=int)
        with pytest.raises(ValueError):
            validate_data_matrix(samples, features, data)


class TestDataMatrix:
    n_samples = 10
    n_features = 20

    def test_create_matrix_using_samples_unsorted_samples_sorts_by_order(self, tmp_path: pathlib.Path):
        samples = list(reversed([create_sample(tmp_path, x) for x in range(3)]))
        features = [create_feature_group(x) for x in range(5)]
        data = numpy.random.normal(loc=100.0, size=(3, 5))
        matrix = DataMatrix(samples, features, data)

        assert matrix.samples == tuple(reversed(samples))
        assert data is not matrix.get_data()
        assert numpy.array_equal(data[::-1], matrix.get_data())

    @pytest.fixture
    def samples(self, tmp_path: pathlib.Path):
        return tuple(create_sample(tmp_path, k) for k in range(self.n_samples))

    @pytest.fixture
    def features(self):
        return tuple(create_feature_group(k) for k in range(self.n_features))

    @pytest.fixture
    def matrix(self, samples, features):
        data = numpy.random.normal(loc=100.0, size=(self.n_samples, self.n_features))
        return DataMatrix(samples, features, data)

    def test_get_n_features(self, matrix):
        assert matrix.get_n_features() == self.n_features

    def test_get_n_samples(self, matrix):
        assert matrix.get_n_samples() == self.n_samples

    def test_list_samples(self, matrix: DataMatrix, samples):
        assert samples == matrix.samples

    def test_list_features(self, matrix, features):
        assert features == matrix.features

    def test_get_columns_missing_group_raises_error(self, matrix: DataMatrix):
        with pytest.raises(exceptions.FeatureGroupNotFound):
            invalid_group = 100000
            matrix.get_columns(invalid_group)

    def test_get_columns_single_column(self, matrix: DataMatrix):
        expected_feature = matrix.features[1]
        column = matrix.get_columns(expected_feature.group)[0]
        assert column.feature == expected_feature
        assert numpy.array_equal(column.data, matrix.get_data()[:, 1])

    def test_get_columns_multiple_columns(self, matrix: DataMatrix):
        index = [1, 3, 4]
        expected_features = matrix.features
        query_groups = [expected_features[x].group for x in index]
        columns = matrix.get_columns(*query_groups)
        for ind, col in zip(index, columns):
            assert col.feature == expected_features[ind]
            assert numpy.array_equal(col.data, matrix.get_data()[:, ind])

    def test_get_columns_no_groups_retrieve_all_columns(self, matrix: DataMatrix):
        expected_features = matrix.features
        columns = matrix.get_columns()
        for col, ft, col_data in zip(columns, expected_features, matrix.get_data().T):
            assert col.feature == ft
            assert numpy.array_equal(col.data, col_data)

    def test_get_samples_missing_sample_id_raises_error(self, matrix: DataMatrix):
        with pytest.raises(exceptions.SampleNotFound):
            invalid_sample_id = "invalid-sample-id"
            matrix.get_rows(invalid_sample_id)

    def test_get_samples_single_sample(self, matrix: DataMatrix):
        expected_sample = matrix.samples[1]
        row = matrix.get_rows(expected_sample.id)[0]
        assert row.sample == expected_sample
        assert numpy.array_equal(row.data, matrix.get_data()[1])

    def test_get_samples_multiple_samples(self, matrix: DataMatrix):
        index = [1, 3, 4]
        expected_samples = matrix.samples
        query_ids = [expected_samples[x].id for x in index]
        rows = matrix.get_rows(*query_ids)
        for ind, row in zip(index, rows):
            assert row.sample == expected_samples[ind]
            assert numpy.array_equal(row.data, matrix.get_data()[ind])

    def test_get_samples_no_ids_retrieves_all_samples(self, matrix: DataMatrix):
        expected_samples = matrix.samples
        rows = matrix.get_rows()
        for row, sample, data_row in zip(rows, expected_samples, matrix.get_data()):
            assert row.sample == sample
            assert numpy.array_equal(row.data, data_row)

    def test_get_data(self, matrix: DataMatrix):
        data = matrix.get_data()
        expected_shape = (matrix.get_n_samples(), matrix.get_n_features())
        assert data.shape == expected_shape

    def test_get_data_sample_subset(self, matrix: DataMatrix):
        all_samples = matrix.samples
        sample_ids = [all_samples[1].id, all_samples[5].id, all_samples[7].id]
        data = matrix.get_data(sample_ids=sample_ids)
        expected_shape = (len(sample_ids), matrix.get_n_features())
        assert data.shape == expected_shape

        for data_row, row_vector in zip(data, matrix.get_rows(*sample_ids)):
            assert numpy.array_equal(data_row, row_vector.data)

    def test_get_data_missing_sample_id_raises_error(self, matrix: DataMatrix):
        sample_ids = ["invalid_sample_id"]
        with pytest.raises(exceptions.SampleNotFound):
            matrix.get_data(sample_ids=sample_ids)

    def test_get_data_feature_subset(self, matrix: DataMatrix):
        all_features = matrix.features
        feature_groups = [all_features[1].group, all_features[5].group, all_features[11].group]
        data = matrix.get_data(feature_groups=feature_groups)
        expected_shape = (matrix.get_n_samples(), len(feature_groups))
        assert data.shape == expected_shape

        for data_column, column_vector in zip(data.T, matrix.get_columns(*feature_groups)):
            assert numpy.array_equal(data_column, column_vector.data)

    def test_get_data_missing_feature_groups_raises_error(self, matrix: DataMatrix):
        feature_groups = [12314]
        with pytest.raises(exceptions.FeatureGroupNotFound):
            matrix.get_data(feature_groups=feature_groups)

    def test_set_data(self, matrix: DataMatrix):
        n_samples = matrix.get_n_samples()
        n_features = matrix.get_n_features()
        expected_new_data = numpy.random.normal(scale=100.0, size=(n_samples, n_features))

        matrix.set_data(expected_new_data)
        assert numpy.array_equal(matrix.get_data(), expected_new_data)

    def test_set_data_invalid_array_raises_error(self, matrix: DataMatrix):
        n_samples = matrix.get_n_samples()
        n_features = matrix.get_n_features() + 1
        expected_new_data = numpy.random.normal(scale=100.0, size=(n_samples, n_features))
        with pytest.raises(ValueError):
            matrix.set_data(expected_new_data)

    def test_set_columns_no_columns_ok(self, matrix: DataMatrix):
        expected = matrix.get_data().copy()
        matrix.set_columns()
        assert numpy.array_equal(matrix.get_data(), expected)

    def test_set_columns_single_column(self, matrix: DataMatrix):
        expected = matrix.get_data().copy()
        group = 1
        column = (group, numpy.random.normal(size=matrix.get_n_samples()))
        idx = matrix.get_feature_index(group)[0]
        expected[:, idx] = column[1]
        matrix.set_columns(column)
        assert numpy.array_equal(matrix.get_data(), expected)

    def test_set_columns_invalid_group_raise_error(self, matrix: DataMatrix):
        column = (1000, numpy.random.normal(size=matrix.get_n_samples()))
        with pytest.raises(exceptions.FeatureGroupNotFound):
            matrix.set_columns(column)

    def test_set_columns_non_matching_size_raise_error(self, matrix: DataMatrix):
        column = (1, numpy.random.normal(size=matrix.get_n_samples() + 1))
        with pytest.raises(ValueError):
            matrix.set_columns(column)

    def test_set_columns_non_float_dtype_raise_error(self, matrix: DataMatrix):
        column = (1, numpy.random.normal(size=matrix.get_n_samples()).astype(int))
        with pytest.raises(TypeError):
            matrix.set_columns(column)

    def test_set_columns_multiple_columns(self, matrix: DataMatrix):
        expected = matrix.get_data().copy()
        columns = list()
        for k in [1, 5, 10]:
            col = (k, numpy.random.normal(size=matrix.get_n_samples()))
            columns.append(col)
            idx = matrix.get_feature_index(k)[0]
            expected[:, idx] = col[1]
        matrix.set_columns(*columns)
        assert numpy.array_equal(matrix.get_data(), expected)

    def test_set_rows_no_rows_ok(self, matrix: DataMatrix):
        expected = matrix.get_data().copy()
        matrix.set_rows()
        assert numpy.array_equal(matrix.get_data(), expected)

    def test_set_row_single_row(self, matrix: DataMatrix):
        expected = matrix.get_data().copy()
        sample_id = matrix.samples[2].id
        row = (sample_id, numpy.random.normal(size=matrix.get_n_features()))
        idx = matrix.get_sample_index(sample_id)[0]
        expected[idx] = row[1]
        matrix.set_rows(row)
        assert numpy.array_equal(matrix.get_data(), expected)

    def test_set_rows_invalid_sample_id_raise_error(self, matrix: DataMatrix):
        row = ("invalid_id", numpy.random.normal(size=matrix.get_n_features()))
        with pytest.raises(exceptions.SampleNotFound):
            matrix.set_rows(row)

    def test_set_rows_non_matching_size_raise_error(self, matrix: DataMatrix):
        row = (matrix.samples[0].id, numpy.random.normal(size=matrix.get_n_features() + 1))
        with pytest.raises(ValueError):
            matrix.set_rows(row)

    def test_set_rows_non_float_dtype_raise_error(self, matrix: DataMatrix):
        row = (matrix.samples[0].id, numpy.random.normal(size=matrix.get_n_features()).astype(int))
        with pytest.raises(TypeError):
            matrix.set_rows(row)

    def test_set_rows_multiple_rows(self, matrix: DataMatrix):
        expected = matrix.get_data().copy()
        rows = list()
        for sample_id in [x.id for x in matrix.samples[::2]]:
            row = (sample_id, numpy.random.normal(size=matrix.get_n_features()))
            rows.append(row)
            idx = matrix.get_sample_index(sample_id)[0]
            expected[idx] = row[1]
        matrix.set_rows(*rows)
        assert numpy.array_equal(matrix.get_data(), expected)

    def test_remove_features_remove_invalid_feature_raises_error(self, matrix: DataMatrix):
        with pytest.raises(exceptions.FeatureGroupNotFound):
            matrix.remove_features(100000)

    def test_remove_features_remove_all_features_raises_error(self, matrix: DataMatrix, features: list[FeatureGroup]):
        all_groups = [x.group for x in features]
        with pytest.raises(exceptions.EmptyDataMatrix):
            matrix.remove_features(*all_groups)

    def test_remove_features_no_features_does_nothing(self, matrix: DataMatrix, features: list[FeatureGroup]):
        matrix.remove_features()
        for ft in features:
            assert matrix.has_feature(ft.group)

    def test_remove_features_remove_single_feature(self, matrix: DataMatrix, features: list[FeatureGroup]):
        rm_group = features[3].group

        data_before_delete = matrix.get_data()
        keep_features_index = matrix.get_feature_index(*(x.group for x in features if x.group != rm_group))

        matrix.remove_features(rm_group)

        # check size
        assert (self.n_features - 1) == matrix.get_n_features()

        # check data order
        assert numpy.array_equal(data_before_delete[:, keep_features_index], matrix.get_data())

        # check that removed feature is no longer stored
        assert not matrix.has_feature(rm_group)

        # check feature order
        remaining_features = [x for x in features if x.group != rm_group]
        assert remaining_features == matrix.features

    def test_remove_features_remove_multiple_features(self, matrix: DataMatrix, features: list[FeatureGroup]):
        rm_groups = [features[3].group, features[2].group]  # using unsorted order to check that it still works
        data_before_delete = matrix.get_data()
        keep_features_index = matrix.get_feature_index(*(x.group for x in features if x.group not in rm_groups))

        matrix.remove_features(*rm_groups)

        # check size
        assert (self.n_features - len(rm_groups)) == matrix.get_n_features()

        # check data order
        assert numpy.array_equal(data_before_delete[:, keep_features_index], matrix.get_data())

        # check that removed features are no longer stored
        for g in rm_groups:
            assert not matrix.has_feature(g)

        # check feature order
        remaining_features = [x for x in features if x.group not in rm_groups]
        assert remaining_features == matrix.features

    def test_remove_samples_remove_invalid_sample_raises_error(self, matrix: DataMatrix):
        with pytest.raises(exceptions.SampleNotFound):
            matrix.remove_samples("invalid_sample_id")

    def test_remove_samples_remove_all_samples_raises_error(self, matrix: DataMatrix, samples: list[Sample]):
        all_sample_ids = [x.id for x in samples]
        with pytest.raises(exceptions.EmptyDataMatrix):
            matrix.remove_samples(*all_sample_ids)

    def test_remove_samples_no_samples_does_nothing(self, matrix: DataMatrix, samples: list[Sample]):
        matrix.remove_samples()
        for sample in samples:
            assert matrix.has_sample(sample.id)

    def test_remove_sample_remove_single_sample(self, matrix: DataMatrix, samples: list[Sample]):
        rm_id = samples[2].id

        data_before_delete = matrix.get_data()
        keep_samples_index = matrix.get_sample_index(*(x.id for x in samples if x.id != rm_id))

        matrix.remove_samples(rm_id)

        # check size
        assert (self.n_samples - 1) == matrix.get_n_samples()

        # check data order
        assert numpy.array_equal(data_before_delete[keep_samples_index], matrix.get_data())

        # check that removed sample is no longer stored
        assert not matrix.has_sample(rm_id)

        # check sample order
        remaining_samples = [x for x in samples if x.id != rm_id]
        assert remaining_samples == matrix.samples

    def test_remove_samples_remove_multiple_samples(self, matrix: DataMatrix, samples: list[Sample]):
        rm_ids = [samples[3].id, samples[2].id]  # using unsorted order to check that it still works
        data_before_delete = matrix.get_data()
        keep_samples_index = matrix.get_sample_index(*(x.id for x in samples if x.id not in rm_ids))

        matrix.remove_samples(*rm_ids)

        # check size
        assert (self.n_samples - len(rm_ids)) == matrix.get_n_samples()

        # check data order
        assert numpy.array_equal(data_before_delete[keep_samples_index], matrix.get_data())

        # check that removed samples are no longer stored
        for id_ in rm_ids:
            assert not matrix.has_sample(id_)

        # check sample order
        remaining_samples = [x for x in samples if x.id not in rm_ids]
        assert remaining_samples == matrix.samples

    def test_combine_single_matrix_return_same_matrix(self, matrix: DataMatrix):
        combined = DataMatrix.combine(matrix)
        assert numpy.array_equal(combined.get_data(), matrix.get_data())
        assert combined.samples == matrix.samples
        assert combined.features == matrix.features

    def test_combine_no_matrices_raises_error(self):
        with pytest.raises(ValueError):
            DataMatrix.combine()

    def test_combine_multiple_matrices(self, matrix: DataMatrix):
        all_samples = matrix.samples
        sample_ids1 = [x.id for x in all_samples[:3]]
        sample_ids2 = [x.id for x in all_samples[3:8]]
        sample_ids3 = [x.id for x in all_samples[8:]]
        sub1 = matrix.create_submatrix(sample_ids=sample_ids1)
        sub2 = matrix.create_submatrix(sample_ids=sample_ids2)
        sub3 = matrix.create_submatrix(sample_ids=sample_ids3)

        combined = DataMatrix.combine(sub1, sub2, sub3)

        assert numpy.array_equal(combined.get_data(), matrix.get_data())
        assert combined.samples == matrix.samples
        assert combined.features == matrix.features

    def test_create_submatrix_no_sample_ids_no_feature_groups_return_equal_matrix(self, matrix: DataMatrix):
        actual = matrix.create_submatrix()

        assert actual.features == matrix.features
        assert actual.samples == matrix.samples
        assert numpy.array_equal(actual.get_data(), matrix.get_data())

    def test_create_submatrix_with_sample_ids(self, matrix: DataMatrix):
        all_samples = matrix.samples
        sample_ids = [all_samples[1].id, all_samples[5].id, all_samples[7].id]
        actual = matrix.create_submatrix(sample_ids=sample_ids)

        sub_matrix_samples = actual.samples
        assert len(sub_matrix_samples) == len(sample_ids)
        assert all(actual.has_sample(x) for x in sample_ids)

        assert actual.features == matrix.features
        assert numpy.array_equal(actual.get_data(), matrix.get_data(sample_ids=sample_ids))

    def test_create_submatrix_has_same_status(self, matrix: DataMatrix):
        matrix.check_status()
        assert matrix.status.missing_imputed
        all_samples = matrix.samples
        sample_ids = [all_samples[1].id, all_samples[5].id, all_samples[7].id]
        actual = matrix.create_submatrix(sample_ids=sample_ids)

        sub_matrix_samples = actual.samples
        assert len(sub_matrix_samples) == len(sample_ids)
        assert all(actual.has_sample(x) for x in sample_ids)
        assert actual.status.missing_imputed
        assert actual.status == matrix.status

        assert actual.features == matrix.features
        assert numpy.array_equal(actual.get_data(), matrix.get_data(sample_ids=sample_ids))

    def test_create_submatrix_with_feature_groups(self, matrix: DataMatrix):
        all_features = matrix.features
        feature_groups = [all_features[1].group, all_features[5].group, all_features[11].group]
        actual = matrix.create_submatrix(feature_groups=feature_groups)

        sub_matrix_feature_groups = actual.features
        assert len(sub_matrix_feature_groups) == len(feature_groups)
        assert all(actual.has_feature(x) for x in feature_groups)

        assert actual.samples == matrix.samples
        assert numpy.array_equal(actual.get_data(), matrix.get_data(feature_groups=feature_groups))

    def test_create_submatrix_with_feature_groups_and_sample_ids(self, matrix: DataMatrix):
        all_features = matrix.features
        feature_groups = [all_features[1].group, all_features[5].group, all_features[11].group]
        all_samples = matrix.samples
        sample_ids = [all_samples[1].id, all_samples[5].id, all_samples[7].id]
        actual = matrix.create_submatrix(sample_ids=sample_ids, feature_groups=feature_groups)

        sub_matrix_samples = actual.samples
        assert len(sub_matrix_samples) == len(sample_ids)
        assert all(actual.has_sample(x) for x in sample_ids)

        sub_matrix_feature_groups = actual.features
        assert len(sub_matrix_feature_groups) == len(feature_groups)
        assert all(actual.has_feature(x) for x in feature_groups)

        expected_data = matrix.get_data(feature_groups=feature_groups, sample_ids=sample_ids)

        assert numpy.array_equal(actual.get_data(), expected_data)

    def test_create_submatrix_with_missing_sample_ids_raise_error(self, matrix: DataMatrix):
        all_samples = matrix.samples
        sample_ids = [all_samples[1].id, all_samples[5].id, "invalid_sample_id"]
        with pytest.raises(exceptions.SampleNotFound):
            matrix.create_submatrix(sample_ids=sample_ids)

    def test_create_submatrix_with_missing_feature_group_raise_error(self, matrix: DataMatrix):
        all_features = matrix.features
        invalid_group = 1203
        feature_groups = [all_features[1].group, all_features[5].group, invalid_group]
        with pytest.raises(exceptions.FeatureGroupNotFound):
            matrix.create_submatrix(feature_groups=feature_groups)


class TestDataMatrixQuery:
    n_samples = 10
    n_features = 20

    @pytest.fixture
    def samples(self, tmp_path: pathlib.Path):
        res = list()
        for k in range(self.n_samples):
            group = "odd" if k % 2 else "even"
            batch = k // 10
            s = create_sample(tmp_path, k, order=k, group=group, batch=batch)
            res.append(s)
        return res

    @pytest.fixture
    def features(self):
        return [create_feature_group(k) for k in range(self.n_features)]

    @pytest.fixture
    def matrix(self, samples, features):
        data = numpy.random.normal(loc=100.0, size=(self.n_samples, self.n_features))
        return DataMatrix(samples, features, data)

    def test_fetch_samples_id_no_filter_return_all_samples(self, matrix: DataMatrix):
        _, sample_ids = matrix.query.fetch_sample_ids()[0]
        assert all(matrix.has_sample(x) for x in sample_ids)

    def test_fetch_samples_invalid_filter_field_raises_error(self, matrix: DataMatrix):
        with pytest.raises(exceptions.SampleMetadataNotFound):
            matrix.query.filter(invalid_meta="invalid_value").fetch_sample_ids()

    def test_fetch_samples_with_equality_filter(self, matrix: DataMatrix):
        select_group = "odd"
        _, sample_ids = matrix.query.filter(group=select_group).fetch_sample_ids()[0]
        assert sample_ids
        for id_ in sample_ids:
            sample = matrix.get_sample(id_)
            assert sample.meta.group == "odd"
            assert sample.meta.batch == 0 or sample.meta.batch == 1
        assert all(matrix.has_sample(x) for x in sample_ids)

    def test_fetch_samples_with_in_filter(self, matrix: DataMatrix):
        select_order = [3, 4, 5]
        _, sample_ids = matrix.query.filter(order=select_order).fetch_sample_ids()[0]
        assert len(sample_ids) == len(select_order)
        for id_ in sample_ids:
            sample = matrix.get_sample(id_)
            assert sample.meta.order in select_order
        assert all(matrix.has_sample(x) for x in sample_ids)

    def test_fetch_samples_with_multiple_filters(self, matrix: DataMatrix):
        select_order = [3, 4, 5]
        select_group = "odd"
        _, sample_ids = matrix.query.filter(order=select_order, group=select_group).fetch_sample_ids()[0]
        assert len(sample_ids) == 2
        for id_ in sample_ids:
            sample = matrix.get_sample(id_)
            assert sample.meta.order == 3 or sample.meta.order == 5
            assert sample.meta.group == select_group

    def test_fetch_samples_invalid_group_by_field_raises_error(self, matrix: DataMatrix):
        with pytest.raises(exceptions.SampleMetadataNotFound):
            matrix.query.group_by("invalid_meta").fetch_sample_ids()

    def test_sample_with_group_by_no_groups_return_all_samples_in_an_empty_group(self, matrix: DataMatrix):
        group, sample_ids = matrix.query.group_by().fetch_sample_ids()[0]
        assert not group
        assert len(sample_ids) == matrix.get_n_samples()
        assert all(matrix.has_sample(x) for x in sample_ids)

    def test_sample_with_group_by_single_field(self, matrix: DataMatrix):
        actual = matrix.query.group_by("group").fetch_sample_ids()
        expected_groups = ["odd", "even"]
        for group, sample_ids in actual:
            assert len(group) == 1
            assert group[0] in expected_groups
            for sample_id in sample_ids:
                sample = matrix.get_sample(sample_id)
                assert sample.meta.group == group[0]

    def test_sample_with_group_by_multiple_fields(self, matrix: DataMatrix):
        actual = matrix.query.group_by("group", "batch").fetch_sample_ids()
        expected_groups = ["odd", "even"]
        expected_batches = [0, 1]
        for group, sample_ids in actual:
            group_group, group_batch = group
            assert len(group) == 2
            assert group_group in expected_groups
            assert group_batch in expected_batches
            for sample_id in sample_ids:
                sample = matrix.get_sample(sample_id)
                assert sample.meta.group == group_group
                assert sample.meta.batch == group_batch

    def test_sample_with_filter_and_group_by(self, matrix: DataMatrix):
        actual = matrix.query.filter(batch=1).group_by("group", "batch").fetch_sample_ids()
        expected_groups = ["odd", "even"]
        expected_batch = 1
        for group, sample_ids in actual:
            group_group, group_batch = group
            assert len(group) == 2
            assert group_group in expected_groups
            assert group_batch == expected_batch
            for sample_id in sample_ids:
                sample = matrix.get_sample(sample_id)
                assert sample.meta.group == group_group
                assert sample.meta.batch == expected_batch


class TestDataMatrixMetrics:
    n_samples = 10
    n_features = 20

    @pytest.fixture
    def samples(self, tmp_path: pathlib.Path):
        res = list()
        for k in range(self.n_samples):
            group = "odd" if k % 2 else "even"
            batch = k // 10
            s = create_sample(tmp_path, k, order=k, group=group, batch=batch)
            s.meta.type = "qc" if k < 5 else "sample"
            res.append(s)
        return res

    @pytest.fixture
    def features(self):
        return [create_feature_group(k) for k in range(self.n_features)]

    @pytest.fixture
    def matrix(self, samples, features):
        data = numpy.random.normal(loc=100.0, size=(self.n_samples, self.n_features))
        return DataMatrix(samples, features, data)

    def test_cv(self, matrix: DataMatrix):
        cv = matrix.metrics.cv()
        assert cv.size == matrix.get_n_features()

    def test_detection_rate(self, matrix: DataMatrix):
        dr = matrix.metrics.detection_rate()
        assert dr.size == matrix.get_n_features()

    def test_dratio(self, matrix: DataMatrix):
        dratio = matrix.metrics.dratio()
        assert dratio.size == matrix.get_n_features()

    def test_dratio_with_groups(self, matrix: DataMatrix):
        dratio = matrix.metrics.dratio(sample_groups=["odd"], qc_groups=["even"])
        assert dratio.size == matrix.get_n_features()

    def test_dratio_no_sample_group_raise_error(self, matrix: DataMatrix):
        with pytest.raises(ValueError):
            matrix.metrics.dratio(sample_groups=["invalid_group"])

    def test_dratio_no_qc_group_raise_error(self, matrix: DataMatrix):
        with pytest.raises(ValueError):
            matrix.metrics.dratio(qc_groups=["invalid_group"])

    def test_pca(self, matrix: DataMatrix):
        scores, loadings, variance = matrix.metrics.pca()
        assert scores.shape[0] == matrix.get_n_samples()
        assert loadings is None
        assert variance is None

    def test_pca_matrix_with_missing_values_raise_error(self, samples: list[Sample], features: list[FeatureGroup]):
        data = numpy.random.normal(size=(len(samples), len(features)))
        data[0, 2] = numpy.nan
        matrix = DataMatrix(samples, features, data)
        with pytest.raises(ValueError):
            matrix.metrics.pca()

    def test_correlation(self, matrix: DataMatrix):
        corr = matrix.metrics.correlation("order")
        assert corr.size == matrix.get_n_features()

    def test_correlation_non_numeric_field_raises_error(self, matrix: DataMatrix):
        with pytest.raises(ValueError):
            matrix.metrics.correlation("group")



from __future__ import annotations

from uuid import UUID

import pytest

from tidyms2.core import models

from ..helpers import ConcreteFeature, ConcreteRoi, create_sample


@pytest.fixture
def expected_descriptor_names():
    return ["area", "height", "custom_descriptor", "mz"]


class TestRoi:
    @pytest.fixture(scope="class")
    def sample(self, tmp_path_factory):
        tmp_path = tmp_path_factory.mktemp("tmp")
        return create_sample(tmp_path, 1)

    def test_id(self, sample):
        roi = ConcreteRoi(sample=sample)
        assert isinstance(roi.id, UUID)

    def test_serialization_deserialization_returns_equal_roi(self, sample):
        expected = ConcreteRoi(data=[1.0, 2.0], sample=sample)
        serialized = expected.to_str()
        actual = ConcreteRoi.from_str(serialized, sample=sample)

        assert actual.data == expected.data
        assert actual.id == expected.id


class TestFeature:
    @pytest.fixture(scope="class")
    def sample(self, tmp_path_factory):
        tmp_path = tmp_path_factory.mktemp("tmp")
        return create_sample(tmp_path, 1)

    @pytest.fixture(scope="class")
    def roi(self, sample) -> ConcreteRoi:
        return ConcreteRoi(data=[1.0, 2.0], sample=sample)

    @pytest.fixture(scope="class")
    def feature(self, roi) -> ConcreteFeature:
        return ConcreteFeature(data_mz=1, roi=roi)

    def test_serialization(self, feature):
        serialized = feature.to_str()
        expected = ConcreteFeature.from_str(serialized, feature.roi, feature.annotation)
        assert expected.data_mz == feature.data_mz
        assert expected.roi == feature.roi
        assert expected.annotation is not None
        assert expected.annotation.id == feature.annotation.id

    def test_mz_equals_get_mz(self, feature):
        assert feature.get("mz") == feature.mz

    def test_area_equals_get_area(self, feature):
        assert feature.get("area") == feature.area

    def test_height_equals_get_height(self, feature):
        assert feature.get("height") == feature.height

    def test_custom_descriptor_equals_get_custom_descriptor(self, feature):
        assert feature.get("custom_descriptor") == feature.custom_descriptor

    def test_descriptor_names_are_feature_attributes(self, feature):
        all_descriptors = feature.descriptor_names()
        all_attr = feature.__dict__
        assert all(x in all_attr for x in all_descriptors)

    def test_describe(self, feature):
        descriptors = feature.describe()
        all_attr = feature.__dict__
        assert all(x in all_attr for x in descriptors)
        assert all(isinstance(x, float) for x in descriptors.values())
        assert feature.mz == descriptors["mz"]
        assert feature.height == descriptors["height"]
        assert feature.area == descriptors["area"]

    def test_is_feature_descriptor_in_valid_range_valid_range(self, feature):
        name = "height"
        value = feature.get(name)
        bounds = {name: (value - 5, value + 5)}
        assert feature.has_descriptors_in_range(**bounds)

    def test_is_feature_descriptor_in_valid_range_invalid_range(self, feature):
        name = "height"
        value = feature.get(name)
        bounds = {name: (value + 5, value + 10)}
        assert not feature.has_descriptors_in_range(**bounds)

    @pytest.fixture(scope="class")
    def ft1(self, roi):
        return ConcreteFeature(roi=roi, data_mz=1)

    @pytest.fixture(scope="class")
    def ft2(self, roi):
        return ConcreteFeature(roi=roi, data_mz=2)

    def test_order_lt(self, ft1, ft2):
        assert ft1 < ft2

    def test_le(self, ft1, ft2):
        assert ft1 <= ft2

    def test_order_ge(self, ft1):
        assert ft1 >= ft1

    def test_order_gt(self, ft1, ft2):
        assert ft2 > ft1

    def test_hash_is_equal_to_id_hash(self, ft1):
        assert hash(ft1) == hash(ft1.id)

    def test_get_invalid_descriptor_raises_error(self, ft1):
        with pytest.raises(ValueError):
            ft1.get("invalid_descriptor")


class TestSample:
    def test_serialization(self, tmp_path):
        sample_id = "my-sample"
        path = tmp_path / sample_id
        expected = models.Sample(id=sample_id, path=path)
        actual = models.Sample(**expected.model_dump())
        assert actual == expected

    def test_serialization_with_extra(self, tmp_path):
        sample_id = "my-sample"
        path = tmp_path / sample_id
        extra = {"extra-field-1": 0.25, "extra-field-2": 3, "extra-field-3": "extra"}
        expected = models.Sample(id=sample_id, path=path, meta=extra)  # type: ignore
        actual = models.Sample(**expected.model_dump())
        assert actual == expected



"""Set the current chemical context."""

from .config import ChemicalContextConfiguration
from .envelope_utils import HomoAtomicEnvelopeCache
from .table import PTABLE, PeriodicTable


class ChemicalContext:
    """Centralize chemical information used in the chem package."""

    def __init__(self, config: ChemicalContextConfiguration):
        self.envelope_cache = HomoAtomicEnvelopeCache()
        if config.custom_abundances is None:
            table = PTABLE
        else:
            table = PeriodicTable(custom_abundances=config.custom_abundances)
        self.table = table


DEFAULT_CONTEXT = ChemicalContext(ChemicalContextConfiguration())



from tidyms2.chem import PTABLE


class TestPeriodicTable:
    def test_get_element_using_symbol(self):
        c = PTABLE.get_element("C")
        assert c.z == 6
        assert c.symbol == "C"

    def test_get_element_using_z(self):
        p = PTABLE.get_element(15)
        assert p.symbol == "P"
        assert p.z == 15

    def test_get_isotope_using_symbol(self):
        cl37 = PTABLE.get_isotope("37Cl")
        assert cl37.a == 37
        assert cl37.symbol == "Cl"

    def test_get_isotope_return_an_isotope_copy(self):
        isotope_str = "37Cl"
        cl37_copy = PTABLE.get_isotope(isotope_str, copy=True)
        cl37 = PTABLE.get_isotope(isotope_str)
        assert cl37.a == cl37_copy.a
        assert cl37.m == cl37_copy.m
        assert cl37.z == cl37_copy.z
        assert cl37 is not cl37_copy


def test_Element_get_monoisotope():
    element = PTABLE.get_element("B")
    monoisotope = element.monoisotope
    assert monoisotope.a == 11


def test_Element_get_mmi():
    element = PTABLE.get_element("B")
    mmi = element.mmi
    assert mmi.a == 10



import numpy
import pytest

from tidyms2.core.dataflow import DataMatrixProcessStatus
from tidyms2.core.enums import MSInstrument, Polarity, SeparationMode
from tidyms2.core.exceptions import EmptyDataMatrix
from tidyms2.core.matrix import DataMatrix, FeatureVector, SampleVector
from tidyms2.core.operators.matrix import ColumnFilter, ColumnTransformer, MatrixTransformer, RowFilter, RowTransformer
from tidyms2.core.utils.numpy import FloatArray

from ...helpers import create_data_matrix

N_SAMPLES = 10
N_FEATURES = 20


@pytest.fixture
def matrix(tmp_path):
    return create_data_matrix(N_SAMPLES, N_FEATURES, tmp_path)


class DummyMatrixTransformer(MatrixTransformer):
    value: float = 10.0
    """Set all values in the data matrix to this value"""

    def get_expected_status_in(self) -> DataMatrixProcessStatus:
        return DataMatrixProcessStatus()

    def get_expected_status_out(self) -> DataMatrixProcessStatus:
        return DataMatrixProcessStatus()

    def _transform_matrix(self, data: DataMatrix) -> FloatArray:
        transformed = data.get_data().copy()
        transformed[:] = self.value
        return transformed

    @classmethod
    def from_defaults(cls, instrument: MSInstrument, separation: SeparationMode, polarity: Polarity):
        return cls()


class DummyRowFilter(RowFilter):
    max_order: int = 5
    """Remove samples if their order is larger than this value"""

    def get_expected_status_in(self) -> DataMatrixProcessStatus:
        return DataMatrixProcessStatus()

    def get_expected_status_out(self) -> DataMatrixProcessStatus:
        return DataMatrixProcessStatus()

    def _create_remove_list(self, data: DataMatrix) -> list[str]:
        return [x.id for x in data.samples if x.meta.order > self.max_order]

    @classmethod
    def from_defaults(cls, instrument: MSInstrument, separation: SeparationMode, polarity: Polarity):
        return cls()


class DummyColumnFilter(ColumnFilter):
    max_group: int = 10
    """Remove features if their group is larger than this value"""

    def get_expected_status_in(self) -> DataMatrixProcessStatus:
        return DataMatrixProcessStatus()

    def get_expected_status_out(self) -> DataMatrixProcessStatus:
        return DataMatrixProcessStatus()

    def _create_remove_list(self, data: DataMatrix) -> list[int]:
        return [x.group for x in data.features if x.group > self.max_group]

    @classmethod
    def from_defaults(cls, instrument: MSInstrument, separation: SeparationMode, polarity: Polarity):
        return cls()


class DummyColumnTransformer(ColumnTransformer):
    odd_value: float = 1.0
    """Value to set on columns with odd group values"""

    even_value: float = 2.0
    """Value to set on columns with even group values"""

    def get_expected_status_in(self) -> DataMatrixProcessStatus:
        return DataMatrixProcessStatus()

    def get_expected_status_out(self) -> DataMatrixProcessStatus:
        return DataMatrixProcessStatus()

    def _transform_column(self, column: FeatureVector) -> FeatureVector:
        value = self.odd_value if column.feature.group % 2 else self.even_value
        size = column.data.size
        return FeatureVector(data=numpy.ones(size, dtype=float) * value, feature=column.feature, index=0)

    @classmethod
    def from_defaults(cls, instrument: MSInstrument, separation: SeparationMode, polarity: Polarity):
        return cls()


class DummyRowTransformer(RowTransformer):
    odd_value: float = 1.0
    """Value to set on rows with odd order values"""

    even_value: float = 2.0
    """Value to set on rows with even order values"""

    def get_expected_status_in(self) -> DataMatrixProcessStatus:
        return DataMatrixProcessStatus()

    def get_expected_status_out(self) -> DataMatrixProcessStatus:
        return DataMatrixProcessStatus()

    def _transform_row(self, row: SampleVector) -> SampleVector:
        value = self.odd_value if row.sample.meta.order % 2 else self.even_value
        size = row.data.size
        return SampleVector(data=numpy.ones(size, dtype=float) * value, sample=row.sample, index=0)

    @classmethod
    def from_defaults(cls, instrument: MSInstrument, separation: SeparationMode, polarity: Polarity):
        return cls()


def test_matrix_transformer(matrix: DataMatrix):
    expected = 100.0
    op = DummyMatrixTransformer(value=expected)
    op.apply(matrix)

    assert numpy.allclose(matrix.get_data(), expected)


def test_row_filter(matrix: DataMatrix):
    max_order = 5
    removed = {x.id for x in matrix.samples if x.meta.order > max_order}
    op = DummyRowFilter(max_order=max_order)

    assert all(matrix.has_sample(x) for x in removed)

    op.apply(matrix)

    assert not any(matrix.has_sample(x) for x in removed)


def test_row_remove_none_ok(matrix: DataMatrix):
    max_order = 10000
    all_samples = {x.id for x in matrix.samples}
    op = DummyRowFilter(max_order=max_order)

    assert all(matrix.has_sample(x) for x in all_samples)

    op.apply(matrix)

    assert all(matrix.has_sample(x) for x in all_samples)


def test_row_filter_remove_all_raises_error(matrix: DataMatrix):
    max_order = -1
    op = DummyRowFilter(max_order=max_order)

    with pytest.raises(EmptyDataMatrix):
        op.apply(matrix)


def test_row_filter_with_exclude_list(matrix: DataMatrix):
    max_order = 5
    exclude = [x.id for x in matrix.samples if x.meta.order > 8]
    removed = {x.id for x in matrix.samples if x.meta.order > max_order}.difference(exclude)
    op = DummyRowFilter(max_order=max_order, exclude=exclude)

    assert all(matrix.has_sample(x) for x in removed)
    assert all(matrix.has_sample(x) for x in exclude)

    op.apply(matrix)

    assert not any(matrix.has_sample(x) for x in removed)
    assert all(matrix.has_sample(x) for x in exclude)


def test_column_filter(matrix: DataMatrix):
    max_group = 5
    removed = {x.group for x in matrix.features if x.group > max_group}
    op = DummyColumnFilter(max_group=max_group)

    assert all(matrix.has_feature(x) for x in removed)

    op.apply(matrix)

    assert not any(matrix.has_feature(x) for x in removed)


def test_column_filter_with_exclude_list(matrix: DataMatrix):
    max_group = 5
    exclude = [x.group for x in matrix.features if x.group > 8]
    removed = {x.group for x in matrix.features if x.group > max_group}.difference(exclude)
    op = DummyColumnFilter(max_group=max_group, exclude=exclude)

    assert all(matrix.has_feature(x) for x in removed)
    assert all(matrix.has_feature(x) for x in exclude)

    op.apply(matrix)

    assert not any(matrix.has_feature(x) for x in removed)
    assert all(matrix.has_feature(x) for x in exclude)


def test_column_filter_remove_none_ok(matrix: DataMatrix):
    max_group = 10000
    all_features = {x.group for x in matrix.features}
    op = DummyRowFilter(max_order=max_group)

    assert all(matrix.has_feature(x) for x in all_features)

    op.apply(matrix)

    assert all(matrix.has_feature(x) for x in all_features)


def test_column_filter_remove_all_raises_error(matrix: DataMatrix):
    max_group = -1
    op = DummyColumnFilter(max_group=max_group)

    with pytest.raises(EmptyDataMatrix):
        op.apply(matrix)


@pytest.mark.parametrize("max_workers", [1, 2])
def test_column_transformer(matrix: DataMatrix, max_workers):
    op = DummyColumnTransformer(max_workers=max_workers)
    op.apply(matrix)

    assert matrix.get_n_features()
    for col in matrix.get_columns():
        if col.feature.group % 2:
            assert numpy.allclose(col.data, op.odd_value)
        else:
            assert numpy.allclose(col.data, op.even_value)


@pytest.mark.parametrize("max_workers", [1, 2])
def test_column_transformer_with_exclude_list(matrix: DataMatrix, max_workers):
    exclude_groups = [x.group for x in matrix.features if x.group < 5]
    op = DummyColumnTransformer(max_workers=max_workers, exclude=exclude_groups)

    excluded_columns = {x.feature.group: x for x in matrix.get_columns(*exclude_groups)}
    op.apply(matrix)

    assert matrix.get_n_features()
    for col in matrix.get_columns():
        if col.feature.group in exclude_groups:
            col_before = excluded_columns[col.feature.group]
            assert numpy.array_equal(col.data, col_before.data)
        elif col.feature.group % 2:
            assert numpy.allclose(col.data, op.odd_value)
        else:
            assert numpy.allclose(col.data, op.even_value)


@pytest.mark.parametrize("max_workers", [1, 2])
def test_row_transformer(matrix: DataMatrix, max_workers):
    op = DummyRowTransformer(max_workers=max_workers)
    op.apply(matrix)

    for row in matrix.get_rows():
        if row.sample.meta.order % 2:
            assert numpy.allclose(row.data, op.odd_value)
        else:
            assert numpy.allclose(row.data, op.even_value)


@pytest.mark.parametrize("max_workers", [1, 2])
def test_row_transformer_with_exclude_list(matrix: DataMatrix, max_workers):
    exclude_ids = [x.id for x in matrix.samples if x.meta.order < 5]
    op = DummyRowTransformer(max_workers=max_workers, exclude=exclude_ids)

    excluded_rows = {x.sample.id: x for x in matrix.get_rows(*exclude_ids)}
    op.apply(matrix)

    for row in matrix.get_rows():
        if row.sample.id in exclude_ids:
            col_before = excluded_rows[row.sample.id]
            assert numpy.array_equal(row.data, col_before.data)
        elif row.sample.meta.order % 2:
            assert numpy.allclose(row.data, op.odd_value)
        else:
            assert numpy.allclose(row.data, op.even_value)



"""Data matrix implementation."""

from __future__ import annotations

from functools import cache
from pathlib import Path
from typing import Self, Sequence

import numpy
import pydantic
from typing_extensions import deprecated

from . import exceptions
from .dataflow import DataMatrixProcessStatus
from .enums import CorrelationMethod, NormalizationMethod, SampleType, ScalingMethod
from .models import FeatureGroup, Sample
from .utils import metrics
from .utils.numpy import FloatArray, FloatArray1D


class DataMatrix:
    """Storage class for matrix data.

    :param samples: the list of samples in the data matrix. Each sample is associated with a matrix row.
    :param features: the list of features in the data matrix. Each feature is associated with a matrix column.
    :param data: A 2D numpy float array with matrix data. The number of rows and columns must match the
        `samples`  and `features` length respectively.
    :param validate: If set to ``True`` will assume that the input data is sanitized. Otherwise, will validate
        and normalize data before creating the data matrix. Set to ``True`` by default.

    """

    class Metrics:
        """Define data matrix metrics computation."""

        def __init__(self, matrix: DataMatrix) -> None:
            self.matrix = matrix

        def cv(self, robust: bool = False) -> FloatArray1D:
            r"""Compute features coefficient of variation (CV).

            .. math::

                \textrm{CV} = \frac{\bar{X}}{S}

            where :math:`S` is the sample standard deviation and :math:`\bar{X}` is the sample mean

            :param robust: If set to ``True`` will use the sample median absolute deviation and
                median instead of the standard deviation and mean.
            :return: a dictionary where the keys are group values for each matrix partition and
                the values are the CV estimation for each feature in the group. NaN values will be
                obtained if all values in the column are zero or NaN. If `robust` is set to ``False``
                and the number less than two values in the column are not NaN, a NaN value will also
                be obtained.
            :raises ValueError: if a sample does not contain a metadata field defined in `groupby`.

            """
            return metrics.cv(self.matrix.get_data(), robust=robust)

        def detection_rate(
            self,
            threshold: float | FloatArray1D = 0.0,
        ) -> FloatArray1D:
            """Compute the detection rate of features (DR)."""
            return metrics.detection_rate(self.matrix.get_data(), threshold=threshold)

        def dratio(
            self,
            sample_groups: list[str] | None = None,
            qc_groups: list[str] | None = None,
            robust: bool = False,
        ) -> FloatArray:
            r"""Compute the D-ratio metric for columns.

            The D-ratio is defined as the quotient between the standard deviation of QC data, or data that
            is expected to exhibit instrumental variation only and the standard deviation of sample data
            or data that presents `biological` variation.

            .. math::

                \textrm{D-Ratio} = \frac{S_{\textrm{QC}}}{S_{\textrm{sample}}}

            where :math:`S_{\textrm{sample}` is the sample standard deviation and :math:`S_{\textrm{QC}` is
            the QC standard deviation.

            a D-ratio of 0.0 means that the technical variance is zero, and all observed variance can be
            attributed to a biological cause. On the other hand, a D-Ratio of 1.0 or larger, means that
            the observed variation is mostly technical.

            NaN values in the sample or QC data will be ignored in the computation of the standard deviation.

            :param sample_groups: a list of sample groups with biological variation. If not provided, uses
                all samples with :term:`sample type` :py:class:`SampleType.SAMPLE`.
            :param qc_groups: a list of sample groups with instrumental variation only. If not provided, uses
                all samples with :term:`sample type` :py:class:`SampleType.TECHNICAL_QC`.
            :param robust: if set to ``True`` estimate the D-ratio using the median absolute deviation instead.
            :return: an 1D array with the D-ratio of each column. Columns with constant sample values will
                result in ``Inf``. If both sample and QC columns are constant, the result will be ``NaN``.
                If `robust` is set to ``True`` and there are less than two non ``NaN`` values in either `Xqc`
                or `Xs` columns, NaN values will also be obtained.
            """
            if sample_groups is None:
                sample_query = self.matrix.query.filter(type=SampleType.SAMPLE)
            else:
                sample_query = self.matrix.query.filter(group=sample_groups)

            sample_query_results = sample_query.fetch_sample_ids()
            if not sample_query_results:
                raise ValueError("No samples found in sample group.")
            sample_ids = sample_query_results[0][1]

            if qc_groups is None:
                qc_query = self.matrix.query.filter(type=SampleType.TECHNICAL_QC)
            else:
                qc_query = self.matrix.query.filter(group=qc_groups)

            qc_sample_query_results = qc_query.fetch_sample_ids()
            if not qc_sample_query_results:
                raise ValueError("No samples found in qc group.")
            qc_sample_ids = qc_sample_query_results[0][1]

            Xs = self.matrix.get_data(sample_ids=sample_ids)
            Xqc = self.matrix.get_data(sample_ids=qc_sample_ids)
            return metrics.dratio(Xs, Xqc, robust=robust)

        def lod(self) -> FloatArray:
            r"""Compute the limit of detection (LOD) using blank samples.

            The limit of detection is defined as:

            .. math::

                \textrm{LOD} = \bar{X}_{\textrm{blank}} + 3 * \bar{S}_{\textrm{blank}}

            where :math:`\bar{X}_{\textrm{blank}}` is the feature mean in the blank samples and
            :math:`\bar{S}_{\textrm{blank}}` is the sample standard deviation of blanks.

            :return: an array with the LOD of each feature. If the LOD cannot be estimated because
                there are no blank samples or the blank contains only missing values, it with
                return zero instead.

            """
            blank_query = self.matrix.query.filter(type=SampleType.SAMPLE).fetch_sample_ids()
            if not blank_query:
                return numpy.zeros(shape=(self.matrix.get_n_features(),), dtype=float)
            blank_samples = blank_query[0][1]
            blank_data = self.matrix.get_data(sample_ids=blank_samples)
            return metrics.lod(blank_data)

        def loq(self) -> FloatArray:
            r"""Compute the limit of quantification (LOQ) using blank samples.

            The limit of quantification is defined as:

            .. math::

                \textrm{LOQ} = \bar{X}_{\textrm{blank}} + 10 * \bar{S}_{\textrm{blank}}

            where :math:`\bar{X}_{\textrm{blank}}` is the feature mean in the blank samples and
            :math:`\bar{S}_{\textrm{blank}}` is the sample standard deviation of blanks.

            :return: an array with the LOQ of each feature. If the LOQ cannot be estimated because
                there are no blank samples or the blank contains only missing values, it with
                return zero instead.

            """
            blank_query = self.matrix.query.filter(type=SampleType.SAMPLE).fetch_sample_ids()
            if not blank_query:
                return numpy.zeros(shape=(self.matrix.get_n_features(),), dtype=float)
            blank_samples = blank_query[0][1]
            blank_data = self.matrix.get_data(sample_ids=blank_samples)
            return metrics.loq(blank_data)

        def pca(
            self,
            *,
            n_components: int = 2,
            normalization: NormalizationMethod | str | None = None,
            scaling: ScalingMethod | str | None = None,
            return_loadings: bool = False,
            return_variance: bool = False,
        ):
            """Compute the PCA scores and loading of the data matrix.

            :param n_components: the number of Principal Components to compute.
            :param scaling: the scaling method applied to `X` columns. Refer to
                :py:class:`~tidyms2.core.enums.ScalingMethod` for the list of available scaling methods. If set to
                ``None``, no scaling is applied.
            :param normalization: One of the available normalization methods. Refer to
                :py:class:`~tidyms2.core.enums.NormalizationMethod` for the list of available normalization methods.
                If set to ``None``, do not perform row normalization.
            :param return_loadings: wether to return the PCA feature loadings.
            :param return_variance: wether to return the PC variance.
            :param kwargs: params passed to :py:func:`~tidyms2.core.matrix.DataMatrix.select_samples` method to
                perform a PCA analysis with a subset of samples.
            :return: A 2D array with PCA scores. If `return_loadings` is set to ``True`` also include a 2D array with
                the PCA loadings. If `return_pc_variance` is set to ``True`` also include a 1D array with PC variances.


            """
            if not self.matrix.get_process_status().missing_imputed:
                raise ValueError("PCA cannot be computed on matrices with NaN values.")

            return metrics.pca(
                self.matrix.get_data(),
                n_components=n_components,
                normalization=normalization,
                scaling=scaling,
                return_loadings=return_loadings,
                return_variance=return_variance,
            )

        def correlation(
            self,
            field: str,
            method: CorrelationMethod | str = CorrelationMethod.PEARSON,
        ) -> FloatArray1D:
            """Compute the correlation coefficient between features and a sample metadata field."""
            X = self.matrix.get_data()
            y = numpy.array(self.matrix.list_sample_field(field))
            if y.dtype.kind not in ["i", "u", "f"]:
                raise ValueError("Sample metadata field must contain numeric data.")
            return metrics.correlation(X, y, method)

    class Query:
        """Query API for selecting sample subsets using sample metadata.

        Note that the `filter` and `group_by` methods are implemented using using pure Python.
        If performance is required, consider using the
        :py:func:`~tidyms2.core.matrix.DataMatrix.Metrics.sql` method, which allows to query
        sample metadata and feature using a DuckDB SQL backend.

        """

        def __init__(self, matrix: DataMatrix):
            self.matrix = matrix
            self._filter_by = None
            self._group_by = None

        def _reset(self):
            self._filter_by = None
            self._group_by = None

        def filter(self, **kwargs) -> Self:
            """Select samples based on metadata fields.

            :param kwargs: key-value pairs used to select samples. Keys must be a
                :py:class:`~tidyms2.core.models.SampleMetadata` field. If a scalar value is passed,
                it is compared for equality with each sample metadata. If an list or tuple is passed,
                then the metadata field is checked for membership in the iterable. If multiple
                key-value pairs are provided, samples must pass checks for all pairs.

            """
            self._filter_by = kwargs
            return self

        def group_by(self, *args: str) -> Self:
            """Group samples based on metadata fields.

            :param args: the list of :py:class:`~tidyms2.core.models.SampleMetadata` fields used to
                create groups.

            """
            self._group_by = args
            return self

        def fetch_sample_ids(self) -> list[tuple[Sequence[str], list[str]]]:
            """Execute a sample query."""
            filter_by = self._filter_by or dict()
            filtered = self._filter_samples(self.matrix.samples, **filter_by)

            group_by = self._group_by or tuple()
            grouped = self._group_by_samples(filtered, *group_by)
            self._reset()
            return [(group, [x.id for x in group_samples]) for group, group_samples in grouped.items()]

        def sql(self, stmt: str):
            """Query data matrix metadata using SQL syntax.

            :param stmt: the SQL statement to query data.

            """
            # TODO: use DuckDB + python
            raise NotImplementedError

        @staticmethod
        def _filter_samples(samples: Sequence[Sample], **kwargs) -> list[Sample]:
            """Select a subset of samples using metadata."""
            filtered = list()
            for sample in samples:
                include = True
                for field, filter_value in kwargs.items():
                    if not hasattr(sample.meta, field):
                        msg = f"Sample id: {sample.id}. Field: {field}"
                        raise exceptions.SampleMetadataNotFound(msg)

                    value = getattr(sample.meta, field, None)

                    if include and isinstance(filter_value, (list, tuple)):
                        include = value in filter_value
                    elif include:
                        include = value == filter_value

                    if not include:
                        break

                if include:
                    filtered.append(sample)
            return filtered

        @staticmethod
        def _group_by_samples(samples: list[Sample], *args) -> dict[Sequence[str], list[Sample]]:
            """Select a subset of samples using metadata."""
            group_to_ids = dict()
            for sample in samples:
                group = list()
                for meta_field in args:
                    if not hasattr(sample.meta, meta_field):
                        msg = f"Sample id: {sample.id}. Field: {meta_field}"
                        raise exceptions.SampleMetadataNotFound(msg)
                    meta_value = getattr(sample.meta, meta_field)

                    group.append(meta_value)
                group = tuple(group)
                group_ids = group_to_ids.setdefault(group, list())
                group_ids.append(sample)
            return group_to_ids

    class IO:
        """Manage export and import of a data matrix in a variety of formats."""

        def __init__(self, matrix: DataMatrix):
            self.matrix = matrix

        def features_to_dict(self) -> dict:
            """Export feature metadata into a dataframe-friendly dictionary format."""
            raise NotImplementedError

        def features_to_csv(self, path: Path) -> None:
            """Write feature metadata into a csv file."""
            raise NotImplementedError

        def matrix_to_dict(self) -> dict:
            """Export data matrix into a dataframe-friendly dictionary format."""
            raise NotImplementedError

        def matrix_to_csv(self, path: Path) -> None:
            """Write data matrix into a csv file."""
            raise NotImplementedError

        def samples_to_dict(self) -> dict:
            """Export sample metadata into a dataframe-friendly dictionary format."""
            raise NotImplementedError

        def samples_to_csv(self, path: Path) -> None:
            """Write sample metadata into a csv file."""
            raise NotImplementedError

        @classmethod
        def from_csv(cls, samples_csv: Path, matrix_csv: Path, features_csv: Path) -> DataMatrix:
            """Create a data matrix instance from csv data."""
            raise NotImplementedError

    def __init__(
        self,
        samples: Sequence[Sample],
        features: Sequence[FeatureGroup],
        data: FloatArray,
        validate: bool = True,
        status: DataMatrixProcessStatus | None = None,
    ):
        self._data = data
        self._samples = tuple(x for x in samples)
        self._features = tuple(x for x in features)
        self._status = status or DataMatrixProcessStatus()
        self._metrics = self.Metrics(self)
        self._query = self.Query(self)
        self._io = self.IO(self)
        if validate:
            self.validate()
            self.check_status()

    @property
    def status(self) -> DataMatrixProcessStatus:
        """Data matrix status getter."""
        return self._status

    @property
    def metrics(self) -> Metrics:
        """Matrix metrics method getter."""
        return self._metrics

    @property
    def io(self) -> IO:
        """Matrix IO methods getter."""
        return self._io

    @property
    def query(self) -> Query:
        """Matrix query methods getter."""
        return self._query

    @property
    def samples(self) -> Sequence[Sample]:
        """The list of samples in the matrix."""
        return self._samples

    @property
    def features(self) -> Sequence[FeatureGroup]:
        """The list of features in the matrix."""
        return self._features

    def get_n_features(self) -> int:
        """Retrieve the number of feature groups in the data matrix."""
        return len(self._features)

    def get_n_samples(self) -> int:
        """Retrieve the number of samples in the data matrix."""
        return len(self._samples)

    def get_process_status(self) -> DataMatrixProcessStatus:
        """Retrieve the current data matrix status."""
        return self._status

    @deprecated("Use the 'samples' property  instead.")
    def list_samples(self) -> Sequence[Sample]:
        """List all samples in the data matrix."""
        return self.samples

    def list_sample_field(self, field: str) -> list:
        """Retrieve the field value from all samples.

        If a sample does not contain the queried field, it returns ``None``.

        :param field: the field name to fetch

        """
        return [getattr(x.meta, field, None) for x in self._samples]

    @deprecated("Use the `features` property instead.")
    def list_features(self) -> Sequence[FeatureGroup]:
        """List all features in the data matrix."""
        return self.features

    def add_columns(self, *columns: FeatureVector) -> None:
        """Add columns to the data matrix.

        :param features: the list of columns to add

        """
        ...

    def check_status(self) -> None:
        """Check and update the data matrix status."""
        self._status.missing_imputed = numpy.all(~numpy.isnan(self.get_data())).item()

    def get_columns(self, *groups: int) -> list[FeatureVector]:
        """Retrieve columns from the data matrix.

        :param groups: the feature groups associated with each column. If no groups are provided then all
            groups are retrieved.

        """
        if not groups:
            groups = tuple(self._feature_index())

        columns = list()
        for g in groups:
            if g not in self._feature_index():
                raise exceptions.FeatureGroupNotFound(g)
            index = self._feature_index()[g]
            row = FeatureVector(data=self._data[:, index], feature=self._features[index], index=index)
            columns.append(row)
        return columns

    def get_rows(self, *ids: str) -> list[SampleVector]:
        """Retrieve rows from the data matrix.

        :param ids: the sample ids associated with each row

        """
        if not ids:
            ids = tuple(self._sample_index())

        rows = list()
        for s in ids:
            if s not in self._sample_index():
                raise exceptions.SampleNotFound(s)
            index = self._sample_index()[s]
            row = SampleVector(data=self._data[index], sample=self._samples[index], index=index)
            rows.append(row)
        return rows

    def get_data(
        self,
        sample_ids: list[str] | None = None,
        feature_groups: list[int] | None = None,
    ) -> FloatArray:
        """Retrieve the matrix data in numpy format.

        Each rows in the array is associated with a sample and each column is associated with a feature.

        :param sample_ids: if provided, return a copy of the data array using the subset of samples provided.
        :param feature_groups: if provided, return a copy of the data array using the subset of feature provided.

        """
        if sample_ids is None and feature_groups is None:
            return self._data
        elif sample_ids is None and feature_groups is not None:
            feature_idx = self.get_feature_index(*feature_groups)
            return self._data[:, feature_idx].copy()
        elif sample_ids is not None and feature_groups is None:
            sample_idx = self.get_sample_index(*sample_ids)
            return self._data[sample_idx].copy()
        else:
            assert feature_groups is not None
            assert sample_ids is not None
            feature_idx = self.get_feature_index(*feature_groups)
            sample_idx = self.get_sample_index(*sample_ids)
            return self._data[sample_idx, feature_idx].copy()

    def get_feature_index(self, *groups: int) -> list[int]:
        """Retrieve the list of indices in the data associated with feature groups.

        :param groups: the list of feature groups to search

        """
        try:
            return [self._feature_index()[x] for x in groups]
        except KeyError as e:
            raise exceptions.FeatureGroupNotFound from e

    def get_sample_index(self, *sample_ids: str) -> list[int]:
        """Retrieve the list of indices in the data associated with samples.

        :param sample_ids: the list of samples to search

        """
        try:
            return [self._sample_index()[x] for x in sample_ids]
        except KeyError as e:
            raise exceptions.SampleNotFound from e

    def get_feature(self, group: int) -> FeatureGroup:
        """Retrieve a group from the data matrix.

        :param group: the group label of the feature to retrieve
        :raise FeatureGroupNotFound: if the feature is not found in the data matrix.

        """
        if not self.has_feature(group):
            raise exceptions.FeatureGroupNotFound(group)
        idx = self._feature_index()[group]
        return self._features[idx]

    def get_sample(self, sample_id: str) -> Sample:
        """Retrieve a sample from the data matrix.

        :param sample_id: the id of the sample to retrieve
        :raises SampleNotFound: if no sample with the provided id exists in the matrix

        """
        if not self.has_sample(sample_id):
            raise exceptions.SampleNotFound(sample_id)
        idx = self._sample_index()[sample_id]
        return self._samples[idx]

    def has_feature(self, group: int) -> bool:
        """Check if a feature group is stored in the matrix.

        :param group: the feature group to check

        """
        return group in self._feature_index()

    def has_sample(self, sample_id: str) -> bool:
        """Check if a sample is stored in the matrix.

        :param sample_id: the sample id to check

        """
        return sample_id in self._sample_index()

    def set_columns(self, *pairs: tuple[int, FloatArray1D]) -> None:
        """Set column values in the data matrix.

        :param pairs: a tuple consisting of in a feature group and the corresponding column data.

        """
        n_rows = self.get_n_samples()
        # TODO: check data non-negative
        for group, col in pairs:
            if not self.has_feature(group):
                raise exceptions.FeatureGroupNotFound(group)

            if not col.size == n_rows:
                raise ValueError("The size of columns must be equal to the number of rows in the matrix.")

            if col.dtype.kind != "f":
                raise TypeError("Data array must be of float dtype.")

        for group, col in pairs:
            ind = self._feature_index()[group]
            self._data[:, ind] = col

    def set_data(self, data: FloatArray) -> None:
        """Set all values in the data matrix."""
        validate_data_matrix(self._samples, self._features, data)
        self._data = data.copy()
        self.check_status()

    def set_rows(self, *pairs: tuple[str, FloatArray1D]) -> None:
        """Set row values in the data matrix.

        :param pairs: a tuple consisting of in a sample id and the corresponding column data.

        """
        # TODO: check data non-negative
        n_cols = self.get_n_features()
        for sample_id, row in pairs:
            if not self.has_sample(sample_id):
                raise exceptions.SampleNotFound(sample_id)

            if not row.size == n_cols:
                raise ValueError("Row size must be equal to the number of columns in the matrix.")

            if row.dtype.kind != "f":
                raise TypeError("Row data array must be of float dtype.")

        for sample_id, row in pairs:
            ind = self._sample_index()[sample_id]
            self._data[ind] = row

    def remove_features(self, *groups: int) -> None:
        """Remove feature groups based on their groups labels.

        :param groups: the group labels to remove

        """
        if not groups:
            return

        for g in groups:
            if g not in self._feature_index():
                raise exceptions.FeatureGroupNotFound(g)

        if len(groups) == self.get_n_features():
            raise exceptions.EmptyDataMatrix("Removing selected features will result in an empty data matrix.")

        self._data = numpy.delete(self._data, groups, axis=1)
        group_set = set(groups)
        self._features = [x for x in self._features if x.group not in group_set]
        self._feature_index.cache_clear()

    def remove_samples(self, *ids: str) -> None:
        """Remove samples with based on their ids.

        :param ids: the list of sample ids to remove

        """
        if not ids:
            return

        indices = list()
        for i in ids:
            if i not in self._sample_index():
                raise exceptions.SampleNotFound(i)
            indices.append(self._sample_index()[i])

        if len(indices) == self.get_n_samples():
            raise exceptions.EmptyDataMatrix("Removing the selected samples will result in an empty data matrix.")

        self._data = numpy.delete(self._data, indices, axis=0)
        id_set = set(ids)
        self._samples = [x for x in self._samples if x.id not in id_set]
        self._sample_index.cache_clear()

    def validate(self) -> None:
        """Perform a sanity check and normalization of the data matrix."""
        validate_data_matrix(self._samples, self._features, self._data)
        samples, data = sort_matrix_rows(self._samples, self._data)
        features, data = sort_matrix_columns(self._features, data)
        self._data = data
        self._samples = samples
        self._features = features

    @classmethod
    def combine(cls, *matrices: Self) -> Self:
        """Combine multiple matrices into a single data matrix.

        All matrices are assumed to have the same feature groups.

        """
        if not matrices:
            raise ValueError("At least one matrix is required to perform matrix join.")
        samples = list()
        for m in matrices:
            samples.extend(m.samples)
        data = numpy.vstack([m.get_data() for m in matrices])
        features = matrices[0].features
        return cls(samples, features, data, validate=True)

    def create_submatrix(
        self,
        sample_ids: list[str] | None = None,
        feature_groups: list[int] | None = None,
    ) -> Self:
        """Create a submatrix using a subset of samples and/or features."""
        if sample_ids is None and feature_groups is None:
            submatrix_samples = self.samples
            submatrix_features = self.features
            data = self.get_data()
        elif sample_ids is None and feature_groups is not None:
            submatrix_samples = self.samples
            submatrix_features = [self.get_feature(x) for x in feature_groups]
            data = self.get_data(feature_groups=feature_groups)
        elif sample_ids is not None and feature_groups is None:
            submatrix_samples = [self.get_sample(x) for x in sample_ids]
            submatrix_features = self.features
            data = self.get_data(sample_ids=sample_ids)
        else:
            assert feature_groups is not None
            assert sample_ids is not None
            submatrix_samples = [self.get_sample(x) for x in sample_ids]
            submatrix_features = [self.get_feature(x) for x in feature_groups]
            data = self.get_data(feature_groups=feature_groups, sample_ids=sample_ids)

        return self.__class__(
            submatrix_samples,
            submatrix_features,
            data,
            validate=False,
            status=self._status.model_copy(),
        )

    @cache
    def _sample_index(self) -> dict[str, int]:
        """Map sample ids to indices in the data matrix rows."""
        return {x.id: k for k, x in enumerate(self._samples)}

    @cache
    def _feature_index(self) -> dict[int, int]:
        """Map feature groups to indices in the data matrix columns."""
        return {x.group: k for k, x in enumerate(self._features)}


def validate_data_matrix(samples: Sequence[Sample], features: Sequence[FeatureGroup], data: FloatArray) -> None:
    r"""Perform sanity check on matrix data.

    :param samples: the data matrix samples
    :param features: the data matrix features
    :param data: the data matrix data
    :raises EmptyDataMatrix: if an empty samples or feature list is provided
    :raises ValueError: if the the data shape is not :math:`n_{samples} \times n_{features}` or if the data
        dtype is not float.
    :raises RepeatedIDError: if samples with repeated id are provided.
    :raises RepeatedSampleOrder: if samples with repeated sample order are provided

    """
    if not samples:
        raise exceptions.EmptyDataMatrix("Data matrix must contain at least one sample.")

    if not features:
        raise exceptions.EmptyDataMatrix("Data matrix must contain at least one feature group.")

    if len(data.shape) != 2:
        raise ValueError("data must be a 2D array.")

    if data.dtype.kind != "f":
        raise ValueError("data dtype must be of floating type.")

    n_rows, n_cols = data.shape
    n_samples = len(samples)
    n_features = len(features)

    if n_samples != n_rows:
        msg = f"The number of samples ({n_samples}) does not match the number of rows in the data ({n_rows})."
        raise ValueError(msg)

    if n_features != n_cols:
        msg = f"The number of features ({n_features}) does not match the number of columns in the data ({n_cols})."
        raise ValueError(msg)

    if len({x.group for x in features}) < n_features:
        msg = "Features must have a unique group label."
        raise exceptions.RepeatedIdError(msg)

    Sample.validate_samples(*samples)


def sort_matrix_columns(
    features: Sequence[FeatureGroup], data: FloatArray
) -> tuple[Sequence[FeatureGroup], FloatArray]:
    """Sort data using feature group label."""
    sorted_index = [k for k, _ in sorted(enumerate(features), key=lambda x: x[1].group)]
    return tuple(features[x] for x in sorted_index), data[:, sorted_index]


def sort_matrix_rows(samples: Sequence[Sample], data: FloatArray) -> tuple[Sequence[Sample], FloatArray]:
    """Sort data using sample run order."""
    sorted_index = [k for k, _ in sorted(enumerate(samples), key=lambda x: x[1].meta.order)]
    return tuple(samples[x] for x in sorted_index), data[sorted_index]


class BaseVector(pydantic.BaseModel):
    """A container for 1D vector data."""

    model_config = pydantic.ConfigDict(validate_assignment=True, arbitrary_types_allowed=True)

    data: FloatArray1D = pydantic.Field(repr=False)
    """The vector data."""

    index: int
    """The data matrix index associated with the vector."""


class FeatureVector(BaseVector):
    """Data matrix column view."""

    feature: FeatureGroup
    """The feature information associated with the matrix column."""


class SampleVector(BaseVector):
    """Data matrix row."""

    sample: Sample
    """The sample associated with the row."""



from typing import Sequence

import pytest

from tidyms2.annotation import mmi_finder
from tidyms2.annotation.annotation_data import AnnotationData
from tidyms2.annotation.config import AnnotatorParameters
from tidyms2.chem import DEFAULT_CONTEXT, PTABLE
from tidyms2.core.models import Sample

from ..helpers import ConcreteFeature, ConcreteRoi, create_sample

BIN_SIZE = 100


class TestSelectTwoIsotopeElements:
    CHNOPS = [PTABLE.get_element(x) for x in "CHNOPS"]

    def test_dm_1_p0_greater_than_pi(self):
        expected = [PTABLE.get_element("C")]
        dm = 1
        res = mmi_finder._select_two_isotope_element(self.CHNOPS, dm)
        assert len(res) == len(expected)
        assert set(res) == set(expected)

    def test_dm_1_no_elements(self):
        elements = [PTABLE.get_element(x) for x in "OPS"]
        dm = 1
        res = mmi_finder._select_two_isotope_element(elements, dm)
        assert len(res) == 0

    def test_dm_1_p0_lower_than_pi(self):
        elements = [PTABLE.get_element(x) for x in ["B", "Li", "O", "P", "S"]]
        expected = [PTABLE.get_element(x) for x in ["B", "Li"]]
        dm = 1
        actual = mmi_finder._select_two_isotope_element(elements, dm)
        assert len(actual) == len(expected)
        assert set(actual) == set(expected)

    def test_dm_1_p0_lower_and_higher_than_pi(self):
        elements = [PTABLE.get_element(x) for x in ["C", "H", "B", "Li", "O", "P", "S"]]
        expected = [PTABLE.get_element(x) for x in ["C", "B", "Li"]]
        dm = 1
        actual = mmi_finder._select_two_isotope_element(elements, dm)
        assert len(actual) == len(expected)
        assert set(actual) == set(expected)

    def test_dm_2_p0_greater_than_pi(self):
        elements = [PTABLE.get_element(x) for x in ["Cl", "H", "N", "O", "P", "S"]]
        expected = [PTABLE.get_element(x) for x in ["Cl"]]
        dm = 2
        actual = mmi_finder._select_two_isotope_element(elements, dm)
        assert len(actual) == len(expected)
        assert set(actual) == set(expected)

    def test_dm_2_no_elements(self):
        elements = [PTABLE.get_element(x) for x in ["O", "P", "S"]]
        dm = 2
        res = mmi_finder._select_two_isotope_element(elements, dm)
        assert len(res) == 0

    def test_dm_2_p0_lower_than_pi(self):
        elements = [PTABLE.get_element(x) for x in ["In", "H", "O", "P", "S"]]
        expected = [PTABLE.get_element(x) for x in ["In"]]
        dm = 2
        res = mmi_finder._select_two_isotope_element(elements, dm)
        assert len(res) == len(expected)
        assert set(res) == set(expected)

    def test_dm_2_p0_lower_and_higher_than_pi(self):
        elements = [PTABLE.get_element(x) for x in ["Cl", "In", "Br", "O", "P", "S"]]
        expected = [PTABLE.get_element(x) for x in ["Br", "In"]]
        dm = 2
        res = mmi_finder._select_two_isotope_element(elements, dm)
        assert len(res) == len(expected)
        assert set(res) == set(expected)


class TestSelectMultipleIsotopeElements:
    def test_ok(self):
        elements = [PTABLE.get_element(x) for x in ["Cl", "H", "N", "O", "P", "S"]]
        expected = [PTABLE.get_element(x) for x in ["O", "S"]]
        res = mmi_finder._select_multiple_isotope_elements(elements)
        assert len(res) == len(expected)
        assert set(res) == set(expected)

    def test_no_elements(self):
        elements = [PTABLE.get_element(x) for x in ["Cl", "H", "N", "P"]]
        expected = []
        res = mmi_finder._select_multiple_isotope_elements(elements)
        assert len(res) == len(expected)
        assert set(res) == set(expected)


@pytest.mark.parametrize(
    "elements,expected",
    [
        [["C", "H", "N", "O", "P", "S"], ["C", "O", "S"]],
        [["C", "H", "N", "O", "P", "S", "Cl", "Li", "Na"], ["C", "O", "S", "Li", "Cl"]],
    ],
)
def test__select_elements(elements, expected):
    element_list = [PTABLE.get_element(x) for x in elements]
    expected_elements = [PTABLE.get_element(x) for x in expected]
    actual_elements = mmi_finder._select_elements(element_list)
    assert sorted(actual_elements, key=lambda x: x.z) == sorted(expected_elements, key=lambda x: x.z)


@pytest.fixture(scope="module")
def config():
    bounds = {"C": (0, 108), "H": (0, 100), "S": (0, 8), "Cl": (0, 2)}
    return AnnotatorParameters(max_M=2000.0, bounds=bounds)


@pytest.fixture(scope="module")
def rules(config: AnnotatorParameters) -> dict:
    return mmi_finder._create_rules_dict(config, BIN_SIZE, DEFAULT_CONTEXT)


def create_feature_list(mz: list[float], sp: list[float], sample: Sample) -> Sequence[ConcreteFeature]:
    features = list()
    for k in range(len(mz)):
        ft = ConcreteFeature(data_mz=mz[k], data_area=sp[k], roi=ConcreteRoi(sample=sample))
        features.append(ft)
    return features


@pytest.fixture(scope="module")
def sample(tmp_path_factory) -> Sample:
    tmp_path = tmp_path_factory.mktemp("data")
    return create_sample(tmp_path, 1)


def test__find_candidates(config, rules, sample):
    # create an m/z and sp list where the monoisotopic m/z is the M1 in the isotopic envelope.

    cl = PTABLE.get_element("Cl")
    dm_cl = cl.isotopes[1].a - cl.mmi.a
    mono_mz = 400.0
    charge = 1
    mono_index = 3
    mz = [100.0, 300.0, mono_mz - dm_cl, mono_mz, 456.0]
    sp = [100.0, 200.0, 500.0, 501.0, 34.0]
    peak_list = create_feature_list(mz, sp, sample)
    monoisotope = peak_list[mono_index]

    # find the rule to search the mmi candidate
    m_bin = int(mono_mz // BIN_SIZE)
    i_rules = rules.get(m_bin)[0]

    data = AnnotationData(peak_list)

    test_candidates = mmi_finder._find_candidate(data, monoisotope, charge, i_rules, config)
    mmi = peak_list[2]
    expected_candidates = [(mmi, 1)]
    assert test_candidates == expected_candidates


def test__find_candidates_multiple_candidates(config, rules, sample):
    # create an m/z and sp list where the monoisotopic m/z is the M1 in the
    # isotopic envelope.
    cl = PTABLE.get_element("Cl")
    dm_cl = cl.isotopes[1].a - cl.mmi.a
    mono_mz = 400.0
    charge = 1
    mono_index = 4
    M01 = mono_mz - dm_cl
    M02 = M01 + 0.00001
    mz = [100.0, 300.0, M01, M02, mono_mz, 456.0]
    sp = [100.0, 200.0, 500.0, 500.5, 501.0, 34.0]
    peak_list = create_feature_list(mz, sp, sample)
    monoisotopologue = peak_list[mono_index]

    # find the rule to search the mmi candidate
    m_bin = int(mono_mz // BIN_SIZE)
    i_rules = rules.get(m_bin)[0]

    data = AnnotationData(peak_list)

    test_candidates = mmi_finder._find_candidate(data, monoisotopologue, charge, i_rules, config)
    expected_candidates = [(peak_list[2], 1), (peak_list[3], 1)]
    assert test_candidates == expected_candidates


def test__find_candidates_no_candidates(config, rules, sample):
    # create an m/z and sp list where the monoisotopic m/z is the M1 in the isotopic envelope.
    mono_mz = 400.0
    charge = 1
    mono_index = 2
    mz = [100.0, 300.0, mono_mz, 456.0]
    sp = [100.0, 200.0, 501.0, 34.0]
    peak_list = create_feature_list(mz, sp, sample)
    monoisotopologue = peak_list[mono_index]

    # find the rule to search the mmi candidate
    m_bin = int(mono_mz // BIN_SIZE)
    i_rules = rules.get(m_bin)[0]

    data = AnnotationData(peak_list)

    test_candidates = mmi_finder._find_candidate(data, monoisotopologue, charge, i_rules, config)
    assert len(test_candidates) == 0


def test_MMIFinder(config, sample):
    finder = mmi_finder.MMIFinder(config, BIN_SIZE, DEFAULT_CONTEXT)

    cl = PTABLE.get_element("Cl")
    dm_cl = cl.isotopes[1].a - cl.mmi.a
    mono_mz = 400.0
    mz = [100.0, 300.0, mono_mz - dm_cl, mono_mz, 456.0]
    sp = [100.0, 200.0, 500.0, 501.0, 34.0]
    peak_list = create_feature_list(mz, sp, sample)
    data = AnnotationData(peak_list)
    monoisotopologue = data.get_monoisotopologue()
    test_mmi_index = finder.find(data)
    expected_mmi_index = [
        (monoisotopologue, 1),
        (monoisotopologue, 2),
        (monoisotopologue, 3),
        (peak_list[2], 1),
    ]
    # check with set because features may be in a different order
    assert set(test_mmi_index) == set(expected_mmi_index)



import pytest

from tidyms2.core.models import MZTrace, Sample
from tidyms2.lcms import operators
from tidyms2.lcms.models import Peak
from tidyms2.storage.memory import OnMemorySampleStorage


@pytest.fixture
def sample(lcms_sample_factory) -> Sample:
    return lcms_sample_factory("sample")


@pytest.fixture
def n_features(lcms_sample_factory) -> int:
    sample_spec = lcms_sample_factory.create_simulated_sample_spec()
    return len(sample_spec.features)


@pytest.fixture
def sample_storage(sample) -> OnMemorySampleStorage[MZTrace, Peak]:
    return OnMemorySampleStorage(sample, MZTrace, Peak)


def test_mz_trace_extractor(sample_storage: OnMemorySampleStorage, n_features: int):
    op = operators.LCTraceExtractor(id="mz-trace-extractor")
    assert not sample_storage.get_n_rois()
    op.apply(sample_storage)
    assert sample_storage.get_n_rois() == n_features


def test_peak_extractor(sample_storage: OnMemorySampleStorage, n_features: int):
    # perform roi extraction
    operators.LCTraceExtractor(id="mz-trace-extractor").apply(sample_storage)

    op = operators.LCPeakExtractor(id="peak-extractor")
    assert not sample_storage.get_n_features()
    op.apply(sample_storage)
    assert sample_storage.get_n_rois() == n_features



import os
import pathlib

import pytest
from numpy.random import seed

from tidyms2.io.datasets import download_dataset
from tidyms2.simulation import lcms


@pytest.fixture(scope="session", autouse=True)
def random_seed():
    seed(1234)
    return


@pytest.fixture(scope="session")
def data_dir():
    return pathlib.Path(__file__).parent / "data"


@pytest.fixture(scope="session")
def raw_data_dir(data_dir: pathlib.Path) -> pathlib.Path:
    raw_data_dir = data_dir / "raw"

    if not raw_data_dir.exists():
        raw_data_dir.mkdir()
        dataset = "test-raw-data"
        download_dataset(dataset, raw_data_dir, token=os.getenv("GH_PAT"))

    gitignore = raw_data_dir / ".gitignore"
    gitignore.write_text("*\n")
    return raw_data_dir


@pytest.fixture(scope="module")
def lcms_adducts():
    formula_list = ["[C54H104O6]+", "[C27H40O2]+", "[C24H26O12]+"]
    rt_list = [10.0, 20.0, 30.0]
    abundances = [1000.0, 2000.0, 3000.0]
    adduct_list = list()
    for rt, abundance, formula in zip(rt_list, abundances, formula_list):
        adduct = lcms.SimulatedLCMSAdductSpec(
            formula=formula,
            rt={"mean": rt},  # type: ignore
            abundance={"mean": abundance},  # type: ignore
            n_isotopologues=2,
        )
        adduct_list.append(adduct)
    return adduct_list


@pytest.fixture(scope="module")
def lcms_sample_factory(lcms_adducts):
    config = lcms.DataAcquisitionSpec(n_scans=40, int_std=0.0)
    return lcms.SimulatedLCMSSampleFactory(data_acquisition=config, adducts=lcms_adducts)



import pytest

from tidyms2.core import dataflow
from tidyms2.core.exceptions import ProcessStatusError


class TestCheckProcessStatus:
    def test_base_status_is_compatible_with_itself(self):
        actual = dataflow.SampleProcessStatus()
        expected = dataflow.SampleProcessStatus()
        dataflow.check_process_status(actual, expected)

    def test_expected_roi_extracted_raises_error_if_false(self):
        actual = dataflow.SampleProcessStatus()
        expected = dataflow.SampleProcessStatus(roi_extracted=True)
        with pytest.raises(ProcessStatusError):
            dataflow.check_process_status(actual, expected)

    def test_expected_feature_extracted_raises_error_if_false(self):
        actual = dataflow.SampleProcessStatus()
        expected = dataflow.SampleProcessStatus(feature_extracted=True)
        with pytest.raises(ProcessStatusError):
            dataflow.check_process_status(actual, expected)

    def test_extra_field_not_in_reference_is_ignored(self):
        actual = dataflow.SampleProcessStatus(extra={"extra_check": True})
        expected = dataflow.SampleProcessStatus()
        dataflow.check_process_status(actual, expected)

    def test_extra_field_in_reference_ok(self):
        actual = dataflow.SampleProcessStatus(extra={"extra_expectation": True})
        expected = dataflow.SampleProcessStatus(extra={"extra_expectation": True})
        dataflow.check_process_status(actual, expected)

    def test_extra_field_in_reference_true_raise_error_if_missing(self):
        actual = dataflow.SampleProcessStatus()
        expected = dataflow.SampleProcessStatus(extra={"extra_expectation": True})
        with pytest.raises(ProcessStatusError):
            dataflow.check_process_status(actual, expected)

    def test_extra_field_in_reference_true_raise_error_if_false(self):
        actual = dataflow.SampleProcessStatus(extra={"extra_expectation": False})
        expected = dataflow.SampleProcessStatus(extra={"extra_expectation": True})
        with pytest.raises(ProcessStatusError):
            dataflow.check_process_status(actual, expected)


class TestUpdateProcessStatus:
    def test_update_two_equal_status_does_not_make_changes(self):
        actual = dataflow.SampleProcessStatus()
        expected = actual.model_copy(deep=True)
        reference = dataflow.SampleProcessStatus()
        dataflow.update_process_status(actual, reference)
        assert actual == expected

    def test_update_single_field(self):
        actual = dataflow.SampleProcessStatus()
        assert not actual.roi_extracted
        assert not actual.feature_extracted
        assert not actual.extra
        reference = dataflow.SampleProcessStatus(roi_extracted=True)
        dataflow.update_process_status(actual, reference)
        assert actual.roi_extracted
        assert not actual.feature_extracted
        assert not actual.extra

    def test_update_multiple_fields(self):
        actual = dataflow.SampleProcessStatus()
        assert not actual.roi_extracted
        assert not actual.feature_extracted
        assert not actual.extra
        reference = dataflow.SampleProcessStatus(roi_extracted=True, feature_extracted=True)
        dataflow.update_process_status(actual, reference)
        assert actual.roi_extracted
        assert actual.feature_extracted
        assert not actual.extra

    def test_update_extra(self):
        actual = dataflow.SampleProcessStatus()
        assert not actual.roi_extracted
        assert not actual.feature_extracted
        assert not actual.extra
        extra_field = "extra_check"
        reference = dataflow.SampleProcessStatus(extra={extra_field: True})
        dataflow.update_process_status(actual, reference)
        assert not actual.roi_extracted
        assert not actual.feature_extracted
        assert actual.get_extra(extra_field)



import pytest

from tidyms2.annotation import envelope_finder as ef
from tidyms2.annotation.annotation_data import AnnotationData
from tidyms2.annotation.config import AnnotatorParameters
from tidyms2.chem import DEFAULT_CONTEXT, PTABLE

from ..helpers import create_features_from_formula, create_sample

# Variable name convention
# m denote nominal mass values
# M denote exact mass values
# p denote abundance


@pytest.fixture(scope="module")
def sample(tmp_path_factory):
    tmp_path = tmp_path_factory.mktemp("data")
    return create_sample(tmp_path, 1)


@pytest.fixture(scope="module")
def formula_groups():
    formulas = {
        "cho": ["C27H34O9", "C62H120O6", "C59H114O6", "C62H120O6", "C56H42O10"],
        "chnops": ["C41H80NO8P", "C54H104O6", "C27H40O2", "C24H26O12", "C55H106O6"],
    }
    return formulas


class TestEnvelopeFinderComponents:
    element_groups = {"CHO": ["C", "H", "O"], "CHNOPS": ["C", "H", "N", "O", "P", "S"]}
    formula_groups = {
        "CHO": ["C27H34O9", "C62H120O6", "C59H114O6", "C62H120O6", "C56H42O10"],
        "CHNOPS": ["C41H80NO8P", "C54H104O6", "C27H40O2", "C24H26O12", "C55H106O6"],
    }

    @pytest.mark.parametrize("group", ["CHO", "CHNOPS"])
    def test__make_exact_mass_difference_bounds(self, group):
        # test bounds for different element combinations
        elements = [PTABLE.get_element(x) for x in self.element_groups[group]]
        bounds = ef._make_exact_mass_difference_bounds(elements, 0.0)
        # m and M are the bounds for each nominal mass increment
        for e in elements:
            m = [i.a - e.mmi.a for i in e.isotopes]
            M = [i.m - e.mmi.m for i in e.isotopes]
            for i, mi in zip(m[1:], M[1:]):
                m_min, m_max = bounds[i]
                assert m_min <= mi
                assert m_max >= mi

    @pytest.mark.parametrize("group", ["CHO", "CHNOPS"])
    def test__get_next_mz_search_interval_mz(self, group, sample):
        elements = [PTABLE.get_element(x) for x in self.element_groups[group]]
        dM_bounds = ef._make_exact_mass_difference_bounds(elements, 0.0)
        # test bounds for different formulas
        for f_str in self.formula_groups[group]:
            feature_list = create_features_from_formula(f_str, sample)
            length = len(feature_list)
            for k in range(1, length - 1):
                k_ft = feature_list[k]
                min_mz, max_mz = ef._get_next_mz_search_interval(feature_list[:k], dM_bounds, 1, 0.005)
                assert (min_mz < k_ft.mz) and (k_ft.mz < max_mz)

    @pytest.mark.parametrize("charge", list(range(1, 6)))
    def test_get_k_bounds_multiple_charges(self, charge, sample):
        formulas = self.formula_groups["CHNOPS"]
        elements = [PTABLE.get_element(x) for x in self.element_groups["CHNOPS"]]
        bounds = ef._make_exact_mass_difference_bounds(elements, 0.0)
        for f_str in formulas:
            features = create_features_from_formula(f"[{f_str}]{charge}+", sample)
            length = len(features)
            for k in range(1, length - 1):
                m_min, m_max = ef._get_next_mz_search_interval(features[:k], bounds, charge, 0.005)
                assert (m_min < features[k].mz) and (features[k].mz < m_max)

    @pytest.mark.parametrize("group,charge", [["CHO", 1], ["CHO", 2], ["CHNOPS", 1], ["CHNOPS", 2]])
    def test__find_envelopes(self, group, charge, sample):
        # test that the function works using as a list m/z values generated from
        # formulas.
        formulas = self.formula_groups[group]
        config = AnnotatorParameters(bounds={x: (0, 10) for x in group}, max_M=1000.0)
        finder = ef.EnvelopeFinder(config, DEFAULT_CONTEXT)
        for f_str in formulas:
            features = create_features_from_formula(f"[{f_str}]{charge}+", sample, n_isotopologues=config.max_length)
            data = AnnotationData(features)
            mmi = data.features[0]
            results = finder.find(data, mmi, charge)
            expected = features
            assert results[0] == expected

    @pytest.mark.parametrize("group", ["CHO", "CHNOPS"])
    def test__find_envelopes_no_charge(self, group, sample):
        # test that the function works using as a list m/z values generated from
        # formulas.
        config = AnnotatorParameters(bounds={x: (0, 10) for x in group}, max_M=1000.0)
        finder = ef.EnvelopeFinder(config, DEFAULT_CONTEXT)
        charge = 0
        for f_str in self.formula_groups[group]:
            features = create_features_from_formula(f_str, sample, config.max_length)
            data = AnnotationData(features)
            mmi = features[0]
            results = finder.find(data, mmi, charge)
            expected = features
            assert results[0] == expected

    @pytest.mark.parametrize("group", ["CHNOPS", "CHO"])
    def test_EnvelopeFinder(self, group, sample):
        formulas = self.formula_groups[group]
        config = AnnotatorParameters(bounds={x: (0, 10) for x in group}, max_M=1000.0)
        envelope_finder = ef.EnvelopeFinder(config, DEFAULT_CONTEXT)
        charge = 1
        for f_str in formulas:
            features = create_features_from_formula(f_str, sample, config.max_length)
            mmi = features[0]
            data = AnnotationData(features)
            results = envelope_finder.find(data, mmi, charge)
            expected = features
            assert len(results) == 1
            assert results[0] == expected



"""Utilities to compute isotopic envelopes."""

from __future__ import annotations

from functools import cache
from typing import Generator, Sequence

import numpy as np
import pydantic
from scipy.stats import multinomial

from ..core.utils.numpy import FloatArray, FloatArray1D, IntArray1D, cartesian_product_from_iterable
from .atoms import Element, Isotope


class Bounds(pydantic.BaseModel):
    """Store lower and upper bounds for mass."""

    lower: float
    upper: float

    @pydantic.model_validator(mode="after")
    def _validate_bounds(self):
        if self.lower >= self.upper:
            raise ValueError("Lower bounds must be lower than the upper bound.")
        return self


class EnvelopeBounds(pydantic.BaseModel):
    """Store exact mass and abundance bounds for an envelope query."""

    M: Bounds
    """Exact mass bounds."""

    p: Bounds
    """Abundance bounds."""

    index: int
    """The index used to generate the bounds."""


class EnvelopeQueryTolerance(pydantic.BaseModel):
    """Store the mass and abundance tolerance."""

    min_M_tol: pydantic.PositiveFloat

    max_M_tol: pydantic.PositiveFloat

    p_tol: float = pydantic.Field(lt=1.0, gt=0.0)

    @pydantic.model_validator(mode="after")
    def _check_mass_tolerance(self):
        if self.min_M_tol > self.max_M_tol:
            raise ValueError("`min_M_tol` must be lower than `max_M_tol`.")
        return self


class EnvelopeQuery(pydantic.BaseModel):
    """Container class for Envelope Queries.

    :param M: exact mass of the envelope
    :param p: abundance of the envelope.

    """

    model_config = pydantic.ConfigDict(arbitrary_types_allowed=True)

    M: Sequence[float]
    p: Sequence[float]

    @pydantic.field_validator("M", mode="before")
    @classmethod
    def _check_sorted_mass(cls, M: list[float]) -> list[float]:
        if sorted(M) != M:
            raise ValueError("Values in M must be sorted.")
        return M

    @pydantic.field_validator("M", "p", mode="before")
    @classmethod
    def _check_all_gt_zero(cls, v: list[float]) -> list[float]:
        if any(x <= 0.0 for x in v):
            raise ValueError(f"All values must be greater than zero. Got {v}")
        return v

    @pydantic.field_validator("p", mode="before")
    @classmethod
    def _normalize_p(cls, p: list[float]) -> list[float]:
        norm = sum(p)
        return [x / norm for x in p]

    def crop(self):
        """Reduce the envelope length by one."""
        self.M = [x for x in self.M[:-1]]
        norm = sum(self.p[:-1])
        self.p = [x / norm for x in self.p[:-1]]

    def get_mmi_mass(self):
        """Retrieve the envelope MMI."""
        return self.M[0]

    def _get_mass_tolerance(self, k: int, tol: EnvelopeQueryTolerance) -> float:
        slope = tol.max_M_tol - tol.min_M_tol
        intercept = tol.min_M_tol
        return (1 - self.p[k]) * slope + intercept

    def get_bounds(self, tol: EnvelopeQueryTolerance, k: int) -> EnvelopeBounds:
        """Compute the bounds for the k-th element of the query."""
        if k > len(self.M):
            raise ValueError("`k` must be lower than the length of the query envelope.")
        m_tol = self._get_mass_tolerance(k, tol)
        Mk = self.M[k]
        pk = self.p[k]
        m_bounds = Bounds(lower=min(Mk - m_tol, 0.0), upper=Mk + m_tol)
        p_bounds = Bounds(lower=min(pk - tol.p_tol, 0.0), upper=pk + tol.p_tol)
        return EnvelopeBounds(M=m_bounds, p=p_bounds, index=k)


class Envelope(pydantic.BaseModel):
    """Store the envelope of molecular species."""

    model_config = pydantic.ConfigDict(arbitrary_types_allowed=True)

    M: FloatArray
    """The exact mass values of the species."""

    p: FloatArray
    """The normalized abundance of the species."""

    include: IntArray1D | None = None
    """A list of row indices to include in the envelope. If set to ``None`` include all rows."""

    def iterate_rows(self) -> Generator[tuple[FloatArray1D, FloatArray1D], None, None]:
        """Iterate over rows of both arrays."""
        indices = np.arange(self.get_n_rows()) if self.include is None else self.include

        for i in indices:
            yield self.M[i], self.p[i]

    def get_n_rows(self) -> int:
        """Retrieve the current number of included rows."""
        return self.M.shape[0] if self.include is None else self.include.size

    def get_M_column(self, k):
        """Fetch the k-th column from the mass array."""
        if self.include is None:
            res = self.M[:, k]
        else:
            res = self.M[self.include, k]
        return res

    def get_p_column(self, k):
        """Fetch the k-th column from the abundance array."""
        if self.include is None:
            res = self.p[:, k]
        else:
            res = self.p[self.include, k]
        return res

    def exclude(self, bounds: EnvelopeBounds) -> None:
        """Remove envelopes outside the provided bounds from included envelopes."""
        include = np.arange(self.get_n_rows()) if self.include is None else self.include.copy()

        Mk = self.get_M_column(bounds.index)
        valid_M = (Mk >= bounds.M.lower) & (Mk <= bounds.M.upper)

        pk = self.get_p_column(bounds.index)
        valid_p = (pk >= bounds.p.lower) & (pk <= bounds.p.upper)
        include = include[valid_p & valid_M]

        self.include = include

    def reset_exclude(self):
        """Include all excluded envelopes."""
        self.include = None

    def crop(self, size: int) -> None:
        """Remove columns from mass and abundance arras starting at `size` and renormalize abundances."""
        if size < 2:
            msg = f"Minimum size is 2. Got {size}"
            raise ValueError(msg)

        if size >= self.M.shape[0]:
            return

        self.M = self.M[:, :size]
        self.p = self.p[:, :size]
        normalization = np.sum(self.p, axis=1)
        normalization = normalization.reshape((normalization.size, 1))
        self.p = self.p / normalization


class HomoAtomicEnvelopeCache:
    """Cache the isotopic envelope of homoatomic species.

    :param table: the periodic table used to fetch element data
    :param length: the length of the envelopes.

    """

    def __init__(self):
        pass

    def get_empty(self, length: int = 10) -> Envelope:
        """Get an empty envelope."""
        return _get_n_isotopes_envelope(0.0, 0, length)

    @cache
    def get_envelope(self, x: Element | Isotope, n: int, length: int = 10) -> Envelope:
        """Retrieve the isotopic envelope from a homoatomic species."""
        if n == 0:
            return _get_n_isotopes_envelope(0.0, 0, length)

        if isinstance(x, Isotope):
            return _get_n_isotopes_envelope(x.m, n, length)

        return _get_n_atoms_envelope(x, n, length)

    def make_envelope_array(self, x: Isotope | Element, n_min: int, n_max: int, length: int = 10) -> Envelope:
        """Create an array of exact mass and abundance for homonuclear formulas.

        :param isotope: Isotope
        :param table: PeriodicTable
        :param n_min: minimum formula coefficient
        :param n_max: maximum formula coefficient
        :max_length: max length of the envelope
        :return: a tuple containing a 2D array of envelope exact masses for each n value and a 2D array of abundances.

        """
        rows = n_max - n_min + 1
        M_arr = np.zeros((rows, length))
        p_arr = np.zeros((rows, length))
        for k in range(n_min, n_max + 1):
            k_env = self.get_envelope(x, k, length)
            M_arr[k - n_min] = k_env.M
            p_arr[k - n_min] = k_env.p
        return Envelope(M=M_arr, p=p_arr)


def combine_envelopes(env1: Envelope, env2: Envelope) -> Envelope:
    """Combine the exact mass and abundance of two envelopes.

    All arrays must be 2-dimensional and have the same shape.

    """
    shape = env1.M.shape
    M = np.zeros(shape, dtype=float)
    p = np.zeros(shape, dtype=float)
    # Ignore zero division errors when normalizing by pk
    with np.errstate(divide="ignore", invalid="ignore"):
        for k in range(shape[1]):
            pk = (env1.p[:, : k + 1] * env2.p[:, k::-1]).sum(axis=1)
            k1 = k + 1
            k2 = k
            Mk = (env1.p[:, :k1] * env1.M[:, :k1] * env2.p[:, k2::-1]) + (
                env1.p[:, :k1] * env2.M[:, k2::-1] * env2.p[:, k2::-1]
            )
            M[:, k] = Mk.sum(axis=1) / pk
            p[:, k] = pk
    np.nan_to_num(M, copy=False)
    return Envelope(M=M, p=p)


def _get_n_atoms_envelope(element: Element, n: int, max_length: int) -> Envelope:
    """Compute the envelope of n atoms.

    aux function to _get_n_atoms_envelope.

    """
    n_isotopes = len(element.isotopes)
    # find combinations of isotopes that sum n
    combinations = _find_n_isotope_combination(n_isotopes, n)

    # find m, M and p for each combination of isotopes
    multinomial_dist = multinomial(n, [x.p for x in element.isotopes])
    m = np.matmul(combinations, [x.a for x in element.isotopes])
    M = np.matmul(combinations, [x.m for x in element.isotopes])
    p = multinomial_dist.pmf(combinations)

    # sort by exact mass
    sorted_index = np.argsort(M)
    m, M, p = m[sorted_index], M[sorted_index], p[sorted_index]

    # merge values with the same nominal mass
    _, first_occurrence = np.unique(m, return_index=True)
    m_unique = np.zeros(max_length, dtype=m.dtype)
    M_unique = np.zeros(max_length, dtype=M.dtype)
    p_unique = np.zeros(max_length, dtype=p.dtype)

    # add the length of m_unique to include all nominal mass values
    n_unique = first_occurrence.size
    first_occurrence = list(first_occurrence)
    first_occurrence.append(m.size)
    m0 = m[0]
    for k in range(max_length):
        if k < n_unique:
            start = first_occurrence[k]
            end = first_occurrence[k + 1]
            mk = m[start]
            i = mk - m0
            if i < max_length:
                m_unique[i] = mk
                pk = np.sum(p[start:end])
                p_unique[i] = pk
                M_unique[i] = np.sum(M[start:end] * p[start:end]) / pk
    p_unique = p_unique / np.sum(p_unique)
    return Envelope(M=M_unique.reshape(1, M_unique.size), p=p_unique.reshape(1, p_unique.size))


def _find_n_isotope_combination(n_isotopes, n):
    """Find combinations of isotopes such that the sum is n.

    aux function to _find_n_atoms_abundances.

    """
    n_ranges = [range(x) for x in ([n + 1] * n_isotopes)]
    combinations = cartesian_product_from_iterable(*n_ranges)
    assert combinations is not None
    valid_combinations = combinations.astype(int).sum(axis=1) == n
    combinations = combinations[valid_combinations, :]
    return combinations


def _get_n_isotopes_envelope(M: float, n: int, length: int) -> Envelope:
    """Create the isotopic envelope for n isotopes.

    :param M: the exact mass of the isotope
    :param n: the number of isotopes in the species
    :param length: the envelope length

    aux function to _get_n_atoms_envelope.

    """
    M_arr = np.zeros((1, length), dtype=float)
    p_arr = np.zeros((1, length), dtype=float)
    M_arr[0, 0] = M * n
    p_arr[0, 0] = 1.0
    return Envelope(M=M_arr, p=p_arr)



"""Abstract assay operators."""

from __future__ import annotations

from abc import abstractmethod

from ..dataflow import AssayProcessStatus
from ..models import AnnotationPatch, DescriptorPatch, FillValue
from ..storage import AssayStorage
from .base import BaseOperator


class AssayOperator(BaseOperator[AssayProcessStatus]):
    """Base operator for sample storage."""

    def apply(self, data: AssayStorage) -> None:
        """Apply the operator function to the data."""
        self.check_status(data.get_process_status())

        if hasattr(self, "pre_apply"):
            self.pre_apply()  # type: ignore

        self._apply_operator(data)

        if hasattr(self, "post_apply"):
            self.post_apply()  # type: ignore

        self.update_status(data.get_process_status())

    @abstractmethod
    def _apply_operator(self, data: AssayStorage) -> None: ...


class AnnotationPatcher(AssayOperator):
    """Patches feature annotation data from an assay.

    Must Implement the compute_patches method, which takes an assay storage and returns
    a list of patches that will be applied to the assay.

    """

    def get_expected_status_in(self) -> AssayProcessStatus:
        """Get expected status of input data."""
        return AssayProcessStatus()

    def get_expected_status_out(self) -> AssayProcessStatus:
        """Get status of output data."""
        return AssayProcessStatus()

    def _apply_operator(self, data: AssayStorage) -> None:
        patches = self.compute_patches(data)
        data.patch_annotations(*patches)

    @abstractmethod
    def compute_patches(self, data: AssayStorage) -> list[AnnotationPatch]:
        """Compute feature descriptor patches."""
        ...


class DescriptorPatcher(AssayOperator):
    """Patches descriptor data from an assay.

    Must Implement the compute_patches method, which takes an assay storage and returns
    a list of patches that will be applied to the assay.

    """

    def get_expected_status_in(self) -> AssayProcessStatus:
        """Get expected status of input data."""
        return AssayProcessStatus()

    def get_expected_status_out(self) -> AssayProcessStatus:
        """Get status of output data."""
        return AssayProcessStatus()

    def _apply_operator(self, data: AssayStorage) -> None:
        patches = self.compute_patches(data)
        data.patch_descriptors(*patches)

    @abstractmethod
    def compute_patches(self, data: AssayStorage) -> list[DescriptorPatch]:
        """Compute feature descriptor patches."""
        ...


class MissingImputer(AssayOperator):
    """Add values that will be used as fill in missing data matrix entries.

    Must Implement the `add_fill_values` method, which takes an assay storage and returns
    a list of fill values.

    """

    def get_expected_status_in(self) -> AssayProcessStatus:
        """Get expected status of input data."""
        return AssayProcessStatus(feature_matched=True)

    def get_expected_status_out(self) -> AssayProcessStatus:
        """Get status of output data."""
        return AssayProcessStatus(missing_imputed=True)

    def _apply_operator(self, data: AssayStorage) -> None:
        fill_values = self.compute_fill_values(data)
        data.add_fill_values(*fill_values)

    @abstractmethod
    def compute_fill_values(self, data: AssayStorage) -> list[FillValue]:
        """Compute fill values in a sample.

        :param sample: the sample to search missing values
        :param groups: the feature groups where the sample did not contribute a feature.

        """
        ...
